from pydantic import BaseModel, ValidationError, Field
from pydantic.fields import ModelField
from .download import download, BaseDataLink, reset
from .upload import upload
from ._core_objects import BaseNode, NodeVar, NodeVar_R, NodeVar_W, NodeVar_RW, StaticVar, BaseData, path as to_path

from typing import TypeVar, Generic, Any, Iterable, Dict, List, Type

class C:
    ATTR = 'attr'
    ITEM = 'item'
    NODE = 'node'
    PATH = 'path'

class MatchError(ValueError):
    ...

def _extract_static(obj, name, field):        
    if C.ATTR in field.field_info.extra:
        if field.field_info.extra.get(C.ITEM, None) is not None:
            raise ValueError(f'{C.ATTR!r} and {C.ITEM!r} cannot be both set, choose one.')                    
        
        attribute = field.field_info.extra[C.ATTR]
        if attribute:
            try:
                val = getattr(obj, attribute)
            except AttributeError:
                raise MatchError(f'{attribute!r} is not an attribute of {obj.__class__.__name__!r}')
        else:
            val = obj
    
    elif C.ITEM in field.field_info.extra:             
        item = field.field_info.extra[C.ITEM]
        try:
            val = obj[item]
        except KeyError:
            raise MatchError(f'{item!r} is not an item of {obj.__class__.__name__!r}')
    else:
        try:
            val = getattr(obj, name)
        except AttributeError:
            raise MatchError(f'{name!r} is not an attribute of {obj.__class__.__name__!r}')        
    
    return val
        
def _extract_node(obj, name, field):
    """ called when a NodeVar is detected in datamodel """
    
    if C.NODE in field.field_info.extra:
        
        node = field.field_info.extra[C.NODE]
        node = to_path(node)    
        if field.field_info.extra.get(C.ATTR, None) is not None:
            raise MatchError(f'node={C.NODE!r} and attr={C.ATTR!r} cannot be both set, choose one.')
        
        if field.field_info.extra.get(C.PATH, None) is not None:
            raise MatchError(f'node={C.NODE!r} and path={C.PATH!r} cannot be both set, choose one.')
        
        if field.field_info.extra.get(C.ITEM, None) is not None:
            raise MatchError(f'node={C.NODE!r} and item={C.ITEM!r} cannot be both set, choose one.')
                                        
        if isinstance(node, BaseNode.Property):
            _, node = node.new(obj)
        elif isinstance(node, str):
            try:
                attr = node
                node = getattr(obj, attr)
            except AttributeError:
                raise MatchError(f'{attr!r} is not a node in {obj.__class__.__name__!r}')
        elif hasattr(node, "__iter__"):
            
            attr = node
            cobj = obj
            path = tuple(attr)
            for a in path[:-1]:
                cobj = getattr(cobj, a)
            try:
                node = getattr(cobj, path[-1])
            except AttributeError:
                raise MatchError(f'{path[-1]!r} is not a node in {obj.__class__.__name__!r} with path {path}')        
        elif not isinstance(node, BaseNode):
            raise MatchError(f'node set in the field is not a node')
    
    elif C.ATTR in field.field_info.extra:
        if field.field_info.extra.get(C.ITEM, None) is not None:
            raise MatchError(f'attr={C.ATTR!r} and item={C.ITEM!r} cannot be both set, choose one.')
        if field.field_info.extra.get(C.PATH, None) is not None:
            raise MatchError(f'attr={C.ATTR!r} and path={C.PATH!r} cannot be both set, choose one.') 
            
        attr = field.field_info.extra[C.ATTR]  
        if attr:            
            try:
                node = getattr(obj, attr)
            except AttributeError:
                raise MatchError(f'{attr!r} is not a node in {obj.__class__.__name__!r}')
        else:
            node = obj        
            
    elif C.PATH in field.field_info.extra:
        if field.field_info.extra.get(C.ITEM, None) is not None:
            raise MatchError(f'path={C.PATH!r} and item={C.ITEM!r} cannot be both set, choose one.')
             
        path = field.field_info.extra[C.PATH]
        path = to_path(path)    
        if path:
            
            if isinstance(path, str):
                path = path.split('.')                
            elif not hasattr(path, "__iter__"):
                raise MatchError(f"expecting string or iterable for path parameter got {path!r}")
                
            cobj = obj
            path = tuple(path)
            try:
                for a in path[:-1]:
                    cobj = getattr(cobj, a)
                node = getattr(cobj, path[-1])
            except AttributeError:
                raise MatchError(f'{path!r} is not a valid in {obj.__class__.__name__!r} with path {path}')                                    
        else:
            node = obj
             
        if not isinstance(node, BaseNode):
            raise MatchError(f'node attribute  {attr!r} is not a node in {obj.__class__.__name__!r}')
    
    elif C.ITEM in field.field_info.extra:
         
        item = field.field_info.extra[C.ITEM]
        try:
            node = obj[item]
        except (KeyError, TypeError):
            raise MatchError(f'{item!r} is not an item in {obj.__class__.__name__!r}')
                             
        if not isinstance(node, BaseNode):
            raise MatchError(f'node item {item!r} is not a node in {obj.__class__.__name__!r}')    
                                            
    else:
        attr = name
        try:
            node = getattr(obj, attr)
        except AttributeError as e:
            raise MatchError(f'{attr!r} is not a node in {obj.__class__.__name__!r}')         
        if not isinstance(node, BaseNode):
            raise MatchError(f'node attribute {attr!r} is not a node in {obj.__class__.__name__!r}')
         
    return node                                
    
class DataLink(BaseDataLink):
    """ Link an object containing nodes, to a :class:`pydantic.BaseModel` 
    
    Args:
        input (Any):  Any object with attributes, expecting that the input contains some 
                      :class:`BaseNode` attributes in its hierarchy
                         
        model (:class:`pydantic.BaseModel`): a data model. Is expecting that the data model structure 
            contains :class:`NodeVar` type hint signature.
            
    Example: 
    
        In the following, it is assumed that the motor1 object has a .stat attribute and the .stat
        object have nodes has `pos_actual`, `vel_actual`, etc ... 
        
        ::
        
            from pydevmgr_core import LocalUtcNode, NodeVar, DataLink
            from pydantic import BaseModel, Field
            
            class MotorStatData(BaseModel):
                # the following nodes will be retrieve from the input object, the name here is the 
                # the attribute of the input object                
                pos_actual:  NodeVar[float] = 0.0  
                vel_actual:  NodeVar[float] = 0.0  
                
                
                
                # also the input object attribute pointing to a node can be changed 
                # with the no de keyword in Field (from pydantic import Field)
                pos: NodeVar[float] = Field(0.0, node='pos_actual')
                vel: NodeVar[float] = Field(0.0, node='vel_actual')
                 
                
            class MotorData(BaseModel):    
                num  : int =1 # other data which are not node related                
                
                # Add the stat Data Model 
                stat : MotorStatData = MotorStatData()
                
                # This node is standalone, not linked to input object, 
                # it must be specified in Field with the node keyword 
                utc:         NodeVar[str]   = Field('1950-01-01T00:00:00.00000', node=LocalUtcNode('utc'))
                
                
            >>> data = MotorData()
            >>> data.stat.pos
            0.0
            >>> link = DataLink( motor1, data )
            >>> link.download() # download node values inside data from tins.motor1
            >>> data.stat.pos
            4.566   
            >>> data.utc
            '2020-12-17T10:05:15.831726'
            
        :class:`DataLink` can be added to a :class:`Downloader`, at init or with the :meth:`Dwonloader.add_datalink` method
        
        ::
        
            from pydevmgr_core import Downloader
            
            >>> downloader = Downloader(link)
            >>> downloader.download()
            
            # Or
            
            >>> connection = downloader.new_connection()
            >>> connection.add_datalink(link)                        
            
    """
    def __init__(self, 
          input : Any, 
          model : BaseModel, 
          strick_match : bool = True 
        ) -> None:
        
        self._rnode_fields = {}
        self._wnode_fields = {}
        if isinstance(input, BaseNode):
            try:
                model.value 
            except AttributeError:
                raise ValueError("To link a Node, the data model must have the 'value' attribute")
            self._collect_single_node(model, input)
            self._rnode_fields[input] = [('value', model)]
            self._wnode_fields[input] = [('value', model)]
            
        else:
            self._collect_nodes(model, input, strick_match)
        self._input = input 
        self._model = model
    
    @property
    def model(self)-> BaseModel:
        return self._model    
    
    @property
    def input(self)-> Any:
        return self._input
    
    @property
    def rnodes(self)-> Iterable:
        return self._rnode_fields
    
    @property
    def wnodes(self)-> Iterable:
        return self._wnode_fields
    
    def _collect_single_node(self, model, input_obj):
        for name, field in model.__fields__.items():
            if not isinstance(field.type_, type):
                continue            
            if issubclass(field.type_, StaticVar):
                val = _extract_static(input_obj, name, field)
                setattr(model, name, val) 
            elif issubclass(field.type_, (NodeVar_R,NodeVar,NodeVar_W, NodeVar_RW)):
                raise ValueError("Linking a single node. Data model should not contain NodeVar fields")
    
    def _collect_nodes(self, model, input_obj, strick_match):
        for name, field in model.__fields__.items():
            if not isinstance(field.type_, type):
                continue        
            if issubclass(field.type_, StaticVar):
                try:
                    val = _extract_static(input_obj, name, field)
                except MatchError as e:
                    if strick_match: raise e
                else:    
                    setattr(model, name, val)
                    
            elif issubclass(field.type_, (NodeVar_R,)):
                try:
                    node = _extract_node(input_obj, name, field)               
                except MatchError as e:
                    if strick_match: raise e
                else:
                    self._rnode_fields.setdefault(node, []).append((name, model))
            
            elif issubclass(field.type_, (NodeVar, NodeVar_RW)):
                try:
                    node = _extract_node(input_obj, name, field)               
                except MatchError as e:
                    if strick_match: raise e
                else:
                    
                    self._rnode_fields.setdefault(node, []).append((name, model))
                    self._wnode_fields.setdefault(node, []).append((name, model))
                       
            elif issubclass(field.type_, (NodeVar_W, )):
                try:
                    node = _extract_node(input_obj, name, field)
                except MatchError as e:
                    if strick_match: raise e
                else:                  
                    self._wnode_fields.setdefault(node, []).append((name, model))   
                
            elif issubclass(field.type_, BaseModel):
                # chidren is ignored if path is broken
                sub_model = getattr(model, name)
                if not isinstance(sub_model, BaseModel):                
                    continue
                
                                                    
                try:
                    sub_obj = _extract_static(input_obj, name, field)
                except MatchError as e:
                    # if BaseData force the existance of the path 
                    if strick_match and issubclass(field.type_, BaseData):
                        raise e
                else:                    
                    self._collect_nodes( sub_model, sub_obj, strick_match)
    
    def download_from_nodes(self, nodevals: Dict[BaseNode,Any]) -> None:
        """ Update the data from a dictionary of node/value pair 
        
        If a node in the dictionary is currently not part of the data model link it is ignored silently 
        """
        for node, val in nodevals.items():
            try:
                lst = self._rnode_fields[node]
            except KeyError:
                pass
            else:
                for attr, obj in lst:
                    setattr(obj, attr, val)
                    
    def _download_from(self, data: dict) -> None:
        for node, lst in self._rnode_fields.items():
            for attr, obj in lst:
                setattr(obj, attr, data[node])
                
    def download(self) -> None:
        """ download Nodes from servers and update the data """
        data = {}
        download( self._rnode_fields, data )
        self._download_from(data)
    
    def reset(self):
        """ All linked nodes with a reset method will be reseted """   
        reset(self._rnode_fields)
        reset(self._wnode_fields)         
                
    def _upload_to(self, todata: dict) -> None:
        for node, lst in self._wnode_fields.items():    
            for attr, obj in lst:
                val = getattr(obj, attr)
                if isinstance(val, BaseNode):
                    continue
                # the last in the list will be set
                # which may not in a hierarchical order 
                # any way several same w-node should be avoided
                todata[node] = val  
                        
    def upload(self) -> None:
        """ upload data value (the one linked to a node) to the server """
        todata = {}
        self._upload_to(todata)
        upload(todata) 

def model_subset(
       class_name: str, 
       Model: Type[BaseModel], 
       fields: List[str], 
       BaseClasses: tuple =(BaseModel,)
    ) -> Type[BaseModel]:
    """ From a pydantic model class create a new model class with a subset of fields 
    
    Args:
        class_name (str): name of the create class model 
        Model (BaseModel): Model root 
        fields (List[str]): list of Model field names 
        BaseClasses (optional, tuple): base classes of the new class default is (BaseModel,)
    """
    annotations = {}
    new_fields = {}
    for field_name in fields:
        field = Model.__fields__[field_name]
        fi = field.field_info
        kwargs = dict(            
            description = fi.description, 
            ge=fi.ge, 
            gt=fi.gt, 
            le=fi.le, 
            lt=fi.lt,
            max_items=fi.max_items, 
            max_length=fi.max_length, 
            min_items=fi.min_items, 
            min_length=fi.min_length, 
            multiple_of=fi.multiple_of, 
            title=fi.title,         
        )
        kwargs.update(fi.extra)
                
        new_field = Field(fi.default, **kwargs)
        annotations[field.name] = field.type_ 
        new_fields[field.name] = new_field
        
    new_class = type(class_name, BaseClasses, new_fields)  
    new_class.__annotations__ = annotations
    return new_class



""" The idea here is to create interface between a data value and a widget 



Three type of interfaces :
    - input : The widget is a text, checkbox, button, etc .. and the data is an input of the application
            It has the methods:
                 set_input(val) :  update the input widget with given value
                 get_input()    :  retrieve the value with the right data type
                 
    - output: The widget is a representation of a value, e.g. QLabel the data is an output of the application 
            It has the methods:
                set_output(val) : update the output widget (e.g. QLabel) with the value of dtype
     
    - input output: Two widgets represent input and output if they differ the output is hilighted in red 
                    This is to be used in a config pannel for instance
            t has the methods:
                 set_input(val)   
                 set_output(val) 
                 get_input()  
                 
    Each of the interface has a variable number of argument and keyword depending on the nature of 
    the data value. 
    For performance purpose The update_* methods are created at frontend in the __init__ 
    function to widgets types. This is to avoid things like isinstance(widget, QLabel) at run time 
    The init shall handle how to update and get widget. So far only used case are set, one can increase 
    the diversity of used widget used.
"""

from collections import OrderedDict
from pydantic import BaseModel
from typing import Callable



class WidgetLink:
    def __init__(self, widget, config):
        self._widget = widget
        self._config = config
    
    def set(self, value):
        raise NotImplementedError("set")
    
    @classmethod
    def accept(cls, widget):
        raise NotImplementedError("accept") 

class Input:
    def __init__(self, widget, config):
        self._widget = widget
        self._config = config
    
    def set(self, value):
        raise NotImplementedError("set")
    
    @classmethod
    def accept(cls, widget):
        raise NotImplementedError("accept") 



def _found_handler(cls, widget):
    for sub in cls.__mro__:
        for key, Obj in sub.__dict__.items():
            if isinstance(Obj, type) and issubclass(Obj, Output):
                if Obj.accept(widget):
                    return Obj
                    
    raise ValueError(f"Cannot found suitable output for class {cls.__name__} and widget {type(widget)}")
            


class _BaseVal_:
    @classmethod
    def parse_config(cls, __config__=None, **kwargs):
        if __config__ is None:
            return cls.Config(**kwargs)
        if isinstance(__config__ , cls.Config):
            return __config__
        if isinstance(__config__, dict):
            return cls.Config( **{**__config__, **kwargs} )
        raise ValueError(f"got an unexpected object for config : {type(__config__)}")



    
class OutputVal(_BaseVal_):
    class Config(BaseModel):
        pass
        
    def __init__(self, output_widget, config=None, **kwargs):        
        self.config = self.parse_config(config, **kwargs)        
        self.set = _found_handler(self.__class__, output_widget)(output_widget, self.config).set
                

class InputVal(_BaseVal_):
    class Config(BaseModel):
        pass

    def __init__(self, input_widget, config=None, **kwargs):
        self.config = self.parse_config(config, **kwargs)
        wh = _found_handler(self.__class__, input_widget)(input_widget, self.config)
        self.set_input = wh.set
        self.get = wh.get
        

class BoolCheck(WidgetLink):
    @classmethod
    def accept(cls, w):
        return hasattr(w, "setChecked")
    def set(self, b):
        return self._widget.setChecked(b)
    def get(self):
        return self._widget.isChecked()
    
class BoolLabel(WidgetLink):
    @classmethod
    def accept(cls, w):
        return hasattr(w, "setText") 
    def set(self, b):
        self._widget.setText( self._config.fmt(b) )     

   



class BoolVal_O(OutputVal):
    class Config(OutputVal.Config):
        fmt: Callable  = staticmethod(lambda v: "[X]" if v else "[ ]")
    
    BoolCheck = BoolCheck
    BoolLabel = BoolLabel



class BoolVal_I(InputVal):
    BoolCheck


            



from .base import BaseUiLinker, record_widget_factory

from PyQt5 import uic
from PyQt5.QtWidgets import QFrame, QWidget, QHBoxLayout, QLabel


class BaseLine(BaseUiLinker):

    class Widget(QWidget):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            ly =  QHBoxLayout()
            txt = QLabel("Base line device. No Widget defined")
            ly.addWidget(txt)
            self.setLayout(ly)


record_widget_factory("line", "Base", BaseLine)




from typing import Any, Optional, Type, Callable, Union, Iterable, List, Generic, TypeVar
from pydantic import BaseModel , create_model, Extra, validator, ValidationError

from pydantic.fields import ModelField
from inspect import signature , _ParameterKind, _empty, isbuiltin
from ._class_recorder import get_parser_class, KINDS, record_class
from ._core_base import reconfig
from .._misc.math_parser import ExpEval
import math
from enum import Enum 

from ..io import load_config
parser_loockup = {}

class _Empty_:
    pass


class AnyParserConfig(BaseModel):
    """ A base model for any kind of parser """
    type: Union[List[Union[str, Callable]], str, Callable] = ""
    class Config:
        extra = Extra.allow

# Config for online built parser
class ParserConfig(BaseModel):
    kind: KINDS = KINDS.PARSER.value
    type: str = ""
    class Config:
        validate_assignment = True
        extra = "forbid"
        
# Config for parser classes 
class ParserElementConfig(BaseModel):    
    kind: KINDS = KINDS.PARSER.value
    type: str = ""
    class Config:
        validate_assignment = True
        extra = "forbid"
        
class BaseParser:       
    """ Callable Parser class """      
    Config = ParserElementConfig
    
    def __init_subclass__(cls, **kwargs) -> None:
        if kwargs:
            cls.Config = create_model(  cls.__name__+"Config",  __base__=cls.Config, **kwargs)
   
    def __init__(self, 
           config: Optional[ParserElementConfig] = None, 
           **kwargs
        ) -> None:
        self.config = reconfig(self.Config, config, kwargs)
    
    @staticmethod
    def parse(value, config: Optional[ParserElementConfig]) -> Any:
        return value 
    
    def __call__(self, value: Any) -> Any:        
        return self.parse(value, self.config)


class _BuiltParser:
    Config = ParserConfig
    __parsers__ = [] # yes this is the list and it is defined in the class 
    
    def __init__(self, config=None, **kwargs):
        self.config = reconfig(self.Config, config, kwargs)
    
    @classmethod
    def parse(cls, value, config):
        for parser in cls.__parsers__:
            value = parser(value, config)
        return value
    
    def __call__(self, value):
        return self.parse(value, self.config)        

    
def to_parser_class(_func_: Callable =None, *, type: Optional[str] = None) -> Type[BaseParser]:
    if _func_ is None:
        def parser_func_decorator(func):
            return _parser_func(func, type)
        return parser_func_decorator
    else:
        return _parser_func(_func_, type)


def _parser_func(func, type_):            
    if not hasattr(func, "__call__"):
        raise ValueError(f"{func} is not callable")
        
    try:    
        s = signature(func)
    except ValueError: # assume it is a builtin class with one argument 
        conf_args = {}
    else:
        posarg = False
        conf_args = {}
        for a,p in s.parameters.items():
            if not posarg: 
                if p.kind!=_ParameterKind.POSITIONAL_OR_KEYWORD:
                    if p.default != _empty:
                        raise ValueError("first function argument should not have default") 
                posarg = True
            else:
                if p.default == _empty:
                    raise ValueError("Only one positional argument accepted")
                if p.annotation == _empty:
                    conf_args[a] = p.default 
                else:
                    conf_args[a] = (p.annotation,p.default)
    
    extras = {}
    if type_ is None:        
        if  "type" in conf_args:
            type_  = conf_args['type']            
        else:
            type_ = func.__name__
            extras['type'] = type_
    else:
        extras['type'] = type_
        
    Config = create_model(type_+"Config", **extras, **conf_args, __base__=BaseParser.Config)
    
    
    if conf_args: 
        conf_args_set = set(conf_args)        
        def parser_func_method(arg, config):
            return func(arg, **{a:getattr(config,a) for a in conf_args_set})
    else:
        def parser_func_method(arg, config):
            return func(arg)
            
    return type(type_+"Parser", (BaseParser,), {'Config':Config, 'parse': staticmethod(parser_func_method)})
    

def func_parser_parse(func):
    def f_parse(value, config):
        return func(value)
    return f_parse



_dynamic_parser_class_loockup = {}
def create_parser_class(
       parsers : Union[Callable, Type[BaseParser], Type[_BuiltParser], BaseParser, str, Iterable]
    ) -> Type[_BuiltParser]:
    """ create a new parser class 
    
    From a list of parser class or instance create a new parser class. The parse function of the new
    class will execute each individual parse of the input classes. 
    The Config class of the new parser class is a combination of all Configs
    
    Args:
        parser: A list of type :class:`BaseParser` or :class:`_BuiltParser` or callable or str (recorded parser name) 
    
    Return:
        Cls (Type[_BuiltParser]): A new parser class 
        
    Example:
    
    ::
    
        >>> from pydevmgr_core import create_parser_class, Clipped, Rounded
        >>> FBR = create_parser_class( (float, Rounded, Clipped) )
        >>> FBR.Config()
        FloatRoundedClippedConfig(kind=<KINDS.PARSER: 'Parser'>, type='Rounded', min=-inf, max=inf, ndigits=0)
        >>> f = FBR( min=0.0, max=9.0, ndigits=2)
        >>> f(5.768976976)
        5.77
        >>> f(12.2323)
        9.0    
    
    ... seealso::
        - :func:`parser`
        - :class:`BaseParser`
        
    """
    
    if isinstance(parsers, str):        
        return get_parser_class(parsers)     
    elif isinstance(parsers, type) and issubclass(parsers, (BaseParser, _BuiltParser)): 
        return parsers
    
    try: # this can be the case if parsers is a type like float
        return get_parser_class(parsers) 
    except (ValueError, TypeError):
        pass
        
    if not hasattr(parsers, "__iter__") or isinstance(parsers, type):
        parsers = [parsers]
    
    # convert str to class before checking in loockup
    parsers = tuple(get_parser_class(p) if isinstance(p,str) else p for p in parsers)
    
    try:
        ParserClass = _dynamic_parser_class_loockup[parsers]
    except KeyError:
        pass
    else:
        return ParserClass
    
    
    _cls_parsers = []
    _model_bases = []
    _name = ""
    _type = ""
    
    
    for obj in parsers:  
        if isinstance(obj, (BaseParser, _BuiltParser)):
            ParserCls = obj.__class__
            _cls_parsers.append( ParserCls.parse )
            Tmp_model = create_model( ParserCls.Config.__name__, **obj.config.dict(exclude_unset=True), __base__= ParserCls.Config)
            _model_bases.append(Tmp_model)
            _name += ParserCls.__name__ + str(id(obj.config))
            _type += str(ParserCls.Config.__fields__['type'].default)
        elif isinstance(obj, type) and issubclass(obj, (BaseParser, _BuiltParser)): 
            ParserCls = obj 
            _cls_parsers.append( ParserCls.parse )                                      
            _model_bases.append(obj.Config)
            _name +=  ParserCls.__name__         
            _type += str(ParserCls.Config.__fields__['type'].default)
            
        elif hasattr(obj, "__call__"):
            _cls_parsers.append( func_parser_parse(obj) )
            _name += getattr(obj, "__name__", "F"+str(id(obj))).capitalize()
            _type += getattr(obj, "__name__", "F"+str(id(obj))).capitalize()
        else:
            raise ValueError(f"Expecting valid parser object got a {type(obj)}")  
                
    if _model_bases:                    
        Config = type(_name+"Config", tuple(_model_bases), {'type':_type})
        ParserClass = type(_name, (_BuiltParser,), {'Config':Config, '__parsers__':_cls_parsers})
    else:
        Config = type(_name+"Config", (ParserConfig,), {'type':_type})
        ParserClass = type(_name, (_BuiltParser,), {'__parsers__':_cls_parsers, 'Config':Config})
    _dynamic_parser_class_loockup[parsers] = ParserClass 
    if _type: 
        record_class(ParserClass, overwrite=True)
    return ParserClass

# def parser(parsers, config=None, **kwargs):
#     Parser = create_parser_class(parsers)
#     return Parser(config=config, **kwargs)        

def parser(parsers, config=None, **kwargs):
    """ return a :class:`BaseParser` object 
    
    Args:
       parsers: :class:`BaseParser` or :class:`_BuiltParser` or :class:`BaseParser.Config`
                or dict or string. 
                Definne the parser class used
       config: optional, configuration of the parser 
       * * kwargs:  Additional key/value configuration pairs 
       
    Exemple::
    
    ::
    
        >>> from pydevmgr_core import parser, Bounded, Rounded
        >>> p = parser( (float,Bounded,Rounded), min=0, max=1.0, ndigits=2)
        >>> p('0.223546243')
        0.22
        >>> p(5.0)
        ValueError: 5.0 is higher than 1.0
        
        Note:
        >>>  p = parser( (float,"Bounded","Rounded"), min=0, max=1.0, ndigits=2)
        # works as well as Bounded and Rounded are recorded classes
        
    ::
       
        >>> from pydevmgr_core import parser, Formula, Clipped
        >>> p = parser( (float, Formula, Clipped), formula="2*x+10", max=100)
        >>> [p(10), p(20), p(30), p(40), p(50), p(60)]
        [30.0, 50.0, 70.0, 90.0, 100.0, 100.0]
        
    
        
    .. seealso::
    
        :func:`create_parser_class`
        
    """ 
    if isinstance(parsers, (BaseParser, _BuiltParser)) and not config and not kwargs:
        return parsers
    
    if isinstance(parsers, AnyParserConfig): # Use for a generic config of parser
        Parser = create_parser_class(parsers.type)
        kwargs = {**kwargs, **parsers.dict(exclude=set(["type"]))}
        
    elif isinstance(parsers, (BaseParser.Config, _BuiltParser.Config)):
        return get_parser_class(parsers.type)(config=parsers, **kwargs)
        
    elif isinstance(parsers, dict):
        type_ = parsers.get('type', None)
        if not type_ :
            raise ValueError('parser type cannot be None')
        Parser = create_parser_class(type_)
        parsers = dict(parsers) # make a copy 
        parsers.pop("type", None) # type shall be inside the class definition (always as a string)
        kwargs = {**parsers, **kwargs}    
    else:    
        Parser = create_parser_class(parsers)
    return Parser(config=config, **kwargs)        


ParserVar = TypeVar('ParserVar')
class _BaseParserTyping(Generic[ParserVar]):
    _parser = None
    @classmethod
    def __get_validators__(cls):
        # one or more validators may be yielded which will be called in the
        # order to validate the input, each validator will receive as an input
        # the value returned from the previous validator
        yield cls.validate

    @classmethod
    def __modify_schema__(cls, field_schema):
        # __modify_schema__ should mutate the dict it receives in place,
        # the returned value will be ignored
        #field_schema.update()
        pass

    @classmethod
    def validate(cls, v, field: ModelField):
        
        errors = []

        if field.sub_fields:
            if len(field.sub_fields)>1:
                raise ValidationError(['to many field Defaults require and accept only one argument'], cls)
            val_f = field.sub_fields[0]
                       
            valid_value, error = val_f.validate(v, {}, loc='value')
            
            if error:    
                errors.append(error)
        else:
            val_f = v

        if errors:
            raise ValidationError(errors, cls)

        if cls._parser:
            valid_value = cls._parser(val_f)

            # try:
            #     valid_value = cls._parser(val_f)
            # except ValueError as er:
            #     errors.append(er)
        else:
            valid_value = val_f
            
        
        return valid_value
    
    def __repr__(self):
        return f'{self.__class__.__name__}({super().__repr__()})'

def conparser(parsers, **kwargs):
     """ Build a typing anotation parser for pydantic Models 
        
     Example: 

     ::
        
        from pydantic import BaseModel
        from pydevmgr_core import conparser

        class Conf(BaseModel):
            x: conparser( [float, "Bounded"], min=0, max=10) = 5.0
        
        

        # ValidationError
        >>> Conf(x=11.0)
        ValidationError: 1 validation error for Conf
        x
          11.0 is higher than 10.0 (type=value_error)
        
     """
     p = parser( parsers, **kwargs)
     return type( p.__class__.__name__+"Type", (_BaseParserTyping, ), {"_parser": p})



def _make_global_parsers():
    """ Build automaticaly some parser from python types """
    for tpe in [int, float, complex, bool, str, tuple, set, list]:        
        Tpe = tpe.__name__.capitalize()
        def parse(value, config, tpe=tpe):
            return tpe(value)
        class Config(BaseParser.Config):
            type: str = Tpe 
        cls = type( Tpe , (BaseParser,), {'parse':staticmethod(parse), 'Config':Config} )    
        record_class(cls)
        record_class(cls, type=tpe.__name__)
        globals()[ Tpe ] = cls
_make_global_parsers()

@record_class  
class Clipped(BaseParser):
    class Config(BaseParser.Config):
        type: str = "Clipped"        
        min: float = -math.inf
        max: float = math.inf
    
    @staticmethod
    def parse(value, config):
        return min(config.max,max(config.min, value))
            
@record_class
class Bounded(BaseParser):
    class Config(BaseParser.Config): 
        type: str = "Bounded" 
        min: float = -math.inf
        max: float = math.inf
    
    @staticmethod
    def parse(value, config):        
        if value<config.min :
            raise ValueError(f'{value} is lower than {config.min}')
        if value>config.max :
            raise ValueError(f'{value} is higher than {config.max}')
        return value

@record_class
class Loockup(BaseParser):
    class Config(BaseParser.Config):
        type: str = "Loockup"     
        loockup : list = []
        loockup_default : Optional[Any] = _Empty_
    
    @staticmethod
    def parse(value, config):
        if value not in config.loockup:
            try:
                if config.loockup_default is not _Empty_:
                    return config.loockup_default
                else:
                    raise ValueError(f'must be one of {config.loockup} got {value}')
            except KeyError:            
                raise ValueError(f'must be one of {config.loockup} got {value}')
        return value    


class _BaseEnum(Enum):
    pass
    
@record_class
class Enumerated(BaseParser):    
    class Config(BaseParser.Config):
        type = "Enumerated"
        enumname: str = "" # name of the Enum class if enumarator is a dictionary 
        enumerator: Type = _BaseEnum 
        
        @validator("enumerator", pre=True, check_fields=False)
        def _enum_validator(cls, value, values):
            if isinstance( value, list):
                value = dict(value)
            if isinstance(value, dict):
                return Enum( values['enumname'] or "TmpEnumerator", value)
            return value 
            
    @staticmethod
    def parse(value, config):
        return config.enumerator(value)    
        
@record_class
class Rounded(BaseParser):
    class Config(BaseParser.Config):
        type: str = "Rounded"      
        ndigits: Optional[int] = 0 
        
    @staticmethod      
    def parse(value, config):
        return round(value, config.ndigits)          

@record_class
class ToString(BaseParser):
    class Config(BaseParser.Config):
        type: str = "ToString"      
        format : str = "%s"
        
    @staticmethod    
    def parse(value, config):
        return config.format%(value,)

@record_class
class Capitalized(BaseParser):
    class Config(BaseParser.Config):
        type: str = "Capitalized" 
    @staticmethod
    def parse(value, config):
        return value.capitalize()

@record_class(type="Lower")
@record_class
class Lowered(BaseParser):
    class Config(BaseParser.Config):
        type: str = "Lowered"
    @staticmethod
    def parse(value, config):
        return value.lower()

@record_class(type="Upper")
@record_class
class Uppered(BaseParser):
    class Config(BaseParser.Config):
        type: str = "Uppered"
    @staticmethod
    def parse(value, config):
        return value.upper()

@record_class
class Stripped(BaseParser):
    class Config(BaseParser.Config):
        type: str = "Stripped"
        strip: Optional[str] = None
    @staticmethod
    def parse(value, config):
        return value.strip(config.strip)

@record_class
class LStripped(BaseParser):
    class Config(BaseParser.Config):
        type: str = "LStripped"
        lstrip: Optional[str] = None
    @staticmethod
    def parse(value, config):
        return value.lstrip(config.lstrip)

@record_class
class RStripped(BaseParser):
    class Config(BaseParser.Config):
        type: str = "RStripped"
        rstrip: Optional[str] = None
    @staticmethod
    def parse(value, config):
        return value.rstrip(config.rstrip)


@record_class
class Formula(BaseParser):
    class Config(BaseParser.Config):
        type: str = "Formula"
        formula: str = 'x'
        varname: str = 'x'
    
    @staticmethod
    def parse(value, config):
        # Cash the Eval expression inside the condig.__dict__
        
        # exp = config.__dict__.setdefault( "__:"+config.formula, ExpEval(config.formula ))
        exp = ExpEval(config.formula )
        return exp.eval( {config.varname:value} ) 
        

        # return ExpEval(config.formula).eval({config.varname:value})            

class _BaseModelConfigParse(BaseModel):
    class Config:
        extra = "allow"
@record_class
class ConfigParse(BaseParser):
    class Config(BaseParser.Config):
        type: str = "ConfigParse"
        model: Optional[Type] = _BaseModelConfigParse
        
    @staticmethod
    def parse(value, config):
        if isinstance( value, str):
            value = load_config( value )
        if not isinstance( value, (BaseModel, dict) ):
            raise ValueError("input shall be a file path, a BaseModel or dictionary ")
        if config.model:
            if value is None:
                return config.model()
            return config.model.parse_obj(value)
        if value is None:
            return {}
        return value





from PyQt5.QtWidgets import QLayout, QBoxLayout, QHBoxLayout,  QGridLayout, QMainWindow, QVBoxLayout, QWidget, QMenu, QAction, qApp
from PyQt5 import QtCore

from .base import WidgetControl
from .base import get_widget_factory, BaseUiLinker, WidgetFactory
from typing import List, Tuple, Optional
from pydevmgr_core import BaseDevice, DataLink
from .base_view import DevicesWidgetLinker



class ConfigManager(DevicesWidgetLinker.Config):
    pass


class ManagerWidget(QMainWindow):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
        top = QWidget()
        bottom = QWidget()
        left = QWidget()
        right = QWidget()
        body = QWidget()
         
        
        top_layout = QHBoxLayout()
        bottom_layout = QHBoxLayout()
        left_layout = QVBoxLayout()
        right_layout = QVBoxLayout()
        body_layout = QVBoxLayout()


        top.setLayout(top_layout)
        bottom.setLayout(bottom_layout)
        left.setLayout(left_layout)
        right.setLayout(right_layout)
        body.setLayout(body_layout)

        
        main_layout = QGridLayout()
        main = QWidget()
        main.setLayout(main_layout)
        
        main_layout.addWidget( top, 0, 0, 1, 3 )
        main_layout.addWidget( left, 1, 0 )
        main_layout.addWidget( bottom, 2, 0, 1, 3 )
        main_layout.addWidget( right, 1, 2)
        main_layout.addWidget( body, 1, 1)


        
        devices =  DevicesWidgetLinker.Widget()
        body_layout.addWidget(devices)
        
            

        self.setCentralWidget(main)
        
        # main_layout = QVBoxLayout()
        # self.setLayout(main_layout)
        # main_layout.addWidget(self.body)
        
        menuBar = self.menuBar()
        menuBar.setNativeMenuBar(False)
        # Creating menus using a QMenu object
        fileMenu = QMenu("&File", self)
        menuBar.addMenu(fileMenu)
        
        self.resize(750, 1000)
                
        exitAct = QAction('&Exit', self)
        exitAct.setShortcut('Ctrl+Q')
        exitAct.setStatusTip('Exit application')
        exitAct.triggered.connect(qApp.quit)
        
        fileMenu.addAction(exitAct)
        
        self.viewMenu = QMenu("&View", self)
        menuBar.addMenu(self.viewMenu)
        
        self.menuBar = menuBar
        
        
        self.top = top 
        self.right = right
        self.left = left 
        self.bottom = bottom 
        self.body = body 
        self.main = main 
        self.devices = devices

        self.top_layout = top_layout 
        self.right_layout = right_layout
        self.left_layout = left_layout 
        self.bottom_layout = bottom_layout 
        self.body_layout = body_layout 
        self.main_layout = main_layout 





class ManagerLinker(BaseUiLinker):
    Config = ConfigManager
    Widget =  ManagerWidget
    _widget_view_ctrl = None
    _widget_view = None
    
    class Data(DevicesWidgetLinker.Data):
        pass
    
   
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    
    def connect(self, downloader, manager, data = None):
        if data is None:
            data = self.new_data()
        
        self.setup_ui(manager, data)
        
        self._multi_view_ctrl = WidgetControl(
                self._multi_view, 
                downloader, 
                self.list_devices(manager), 
                data
            )
        
        self._multi_view._link(downloader)
        self._link(downloader, DataLink(manager, data), data)
        return self._multi_view_ctrl
    
    def view_changed(self, view_name):
        if self._multi_view_ctrl:
            self._multi_view_ctrl.data.current_view = view_name
            self._multi_view_ctrl.disable()
            self._multi_view_ctrl.enable()
    
    def list_devices(self, manager):
        return list(manager.find( BaseDevice ))

    
    def setup_ui(self, manager, data):
        
        multi_view = DevicesWidgetLinker(widget=self.widget.devices, config=self.config)
        
        multi_view.setup_ui(self.list_devices(manager), data)
        
        self._multi_view = multi_view
                        
        for view_name in self.config.views:
            viewAct = QAction(view_name, self.widget)
            self.widget.viewMenu.addAction(viewAct)

            self.actions.add(self.view_changed, [view_name], feedback=self.feedback).connect_action(viewAct)
                 

    






from ._class_recorder import get_node_class, KINDS
from ._core_base import (_BaseObject, _BaseProperty, kjoin, 
                           ksplit, BaseData
                        )
from ._core_parser import parser, BaseParser, _BuiltParser, AnyParserConfig, create_parser_class
from ._core_obj_dict import ObjDict

from .. import io 

import weakref
from inspect import signature , _empty

from pydantic import create_model, Extra, BaseModel, validator
from typing import Dict, Any, Optional, Union, Callable, Iterable, List, Dict, Tuple



        
class BaseNodeConfig(_BaseObject.Config):
    """ Config for a Node """
    kind: KINDS = KINDS.NODE.value
    type: str = ""
    #parser: Optional[Union[str, Callable, Iterable]] = None
    parser: Optional[Union[AnyParserConfig, str, List[Union[str, Callable]],  Callable]] = None
    
    description: str = ""
    unit: str = ""
    
    @validator('parser')
    def _parser_validator(cls, parsers):
        if parsers is not None:
            return parser(parsers).config
    
    @classmethod
    def validate_kind(cls, kind):
        if kind:
            if kind!=KINDS.NODE:
                raise ValueError(f'expecting a Node kind, got a {kind!r}')
        return kind    

        
class BaseReadCollector:
    """ Object used to collect nodes values from the same server in one roundtrip 
    
    It has two methods :
     .add(node) to add a node to the queue
     .read(data) to read nodes value inside a dictionary (data) 
    """
    ####
    #@ The Read Collector shall collect all nodes having the same sid and read them in one call
    #@ 
    #@ - __init__ : takes as many argument as necessary shall be build by the the BaseNode.read_collector method
    #@ - add : take one argument, the Node. Should add node in the read queue 
    #@ - read : takes a dictionary as arguement, read the nodes and feed the data dictionary where key  
    #@          is the node itself
    #@ The BaseReadCollector is just a dummy implementation where nodes are red one after the other     
    def __init__(self):
        self._nodes = set()
    def add(self, node):
        self._nodes.add(node)
    def read(self, data):
        for node in self._nodes:
            data[node] = node.get()

class BaseWriteCollector:
    """ Object used to write nodes values from the same server in one roundtrip 
    
    Its has two methods :
     .add(node, value) to add a node and its associated value to the queue
     .write() to write (upload) nodes values
    """
    ####
    #@ The Write Collector shall collect all nodes having the same sid, its value, and write them in one call
    #@ 
    #@ - __init__ : takes has many argument has necessary it shall be built by the BaseNode.write_collector method
    #@ - add : take two argument, the Node and its value attached. Should add node,value in the write queue 
    #@ - write  : takes no arguement, write the node/value 
    #@ 
    #@ The BaseWriteCollector is just a dummy implementation where nodes are written one after the other  
    
    def __init__(self):
        self._nodes = {}
    
    def add(self, node, value):
        self._nodes[node] = value
    
    def write(self):
        for node, val  in self._nodes.items():
            node.set(val)

class DictReadCollector:
    """ A collector to read from a dictionary instead of getting node from server 
    
    E.g. to be used as simulator
    """
    def __init__(self, data):
        self._data = data
        self._nodes = set()
        
    def add(self, node):
        self._nodes.add(node)
        
    def read(self, data):
        for node in self._nodes:
            data[node] = self._data[node.key]

class DictWriteCollector:
    """ A collector to write to a dictionary instead of setting node to server 
    
    E.g. to be used as simulator
    """
    def __init__(self, data):
        self._data = data
        self._nodes = {}
        
    def add(self, node, value):
        self._nodes[node] = value
        
    def write(self):
        for node, val  in self._nodes.items():
            self._data[node.key] = val



class NodeProperty(_BaseProperty):    
    fget = None
    fset = None
    
    def getter(self, func):
        """ decoraotr to define the fget function """
        self.fget = func
        return self # must return self
    
    def setter(self, func):
        """ decoraotr to define the fset function """
        self.fset = func    
        return self # must return self
     
    
    def _finalise(self, parent, node):
        # overwrite the fget, fset function to the node if defined 
        parent_wr = weakref.ref(parent) # shall we keep weakref ? 
        if self.fget:            
            def fget(*args, **kwargs):
                return self.fget(parent_wr(), *args, **kwargs)
            node.fget = fget
        if self.fset:
            def fset(*args, **kwargs):
                return self.fset(parent_wr(), *args, **kwargs)
            node.fset = fset
    
    def __call__(self, func):
        """ fget decorator """
        #### This allows the feature:
        #@ @BaseNode.prop('temp')
        #@ def temp(self):
        #@    return ...
        self.fget = func
        return self

class NodeDict(ObjDict):
    class Config(ObjDict.Config):
        kinds = set([KINDS.NODE])
        default_kind = KINDS.NODE

class BaseNode(_BaseObject):
    """ This a base class defining the base methods for a node 
    
    The requirement for a Node is to have: 
        - a .key (str) attribute
        - a .sid (any hashable) attribute (iddenify an id to group nodes together for efficient reading and writting 
            in one server call). If sid is None the node is threated of an "alias Node".
            The only requirement for sid is that it should be hashable.
        - a .get(data=None) method 
        - a .set(value, data=None) method 
        - read_collector() : a constructor for a node collector for reading 
        - write_collector() : a constructor for a node collector for writting
    
    To implement from BaseNode one need to implement the .fget and .fset method (they are called by .get and .set)
    """
    Config = BaseNodeConfig
    Property = NodeProperty

    Dict = NodeDict

    class Data(BaseData):
        value: Any = None    
    _parser = None
    def __init__(self, 
           key: Optional[str] = None, 
           config: Optional[Config] = None,           
           **kwargs
         ) -> None:
                                         
        super().__init__(key, config=config, **kwargs)
        if self._config.parser:
            self._parser = parser(self._config.parser)
            # write in the __dict__ to avoid attribute assignement 
            self._config.__dict__['parser'] = self._parser.config
            

    @property
    def sid(self):
        """ default id server is 0 
        
        The sid property shall be adujsted according to read_collector and write_collector methods 
        """
        return 0
    
    @property
    def parser(self):
        return self._parser
    
    def parse(self, value):
        """ Parse the value as it is done before being treated by the set method 
        
        The parsed value is the value received by the .fset(parsed_value) method 
        if self.parser is None :  `node.parse(value) is value`
        
        Args:
           value (any): value to be parsed
        
        Return:
            parsed_value (any): value parsed as it is done by the .set method 
        """
        if not self._parser:
            return value
        return self._parser(value)
    
                    
    def read_collector(self) -> BaseReadCollector:
        """ return a collector of nodes for readding 
        
        This is used to collect all nodes with the same sid. 
        The result is done in one call per server when ever it is possible. 
        
        The returned object must have:
            a  ``.add(node)`` method 
            a  ``.read(data)`` method 
        
        The BaseReadCollector is however gettting the node value one by one. The method has to be 
        implemented for other Nodes
        """
        return BaseReadCollector()
    
    def write_collector(self) -> BaseWriteCollector:
        """ return a collector of nodes for writting
        
        This is used to collect all nodes with the same sid. All node could be written in one call
        on the server. 
        
        The returned object must have:
            a  ``.add(node, value)`` method 
            a  ``.write()`` method 
        
        The BaseWriteCollector is however setting the node value one by one. The method has to be 
        implemented for other Nodes
        """
        return BaseWriteCollector()
            
    def get(self, data: Dict =None) -> Any:
        """ get value of the data Node 
        
        If the optional data dictionary is given data[self] is return otherwise self.fget() is returned
        fget() will fetch the value from a distant server for instance (OPC-UA, Websocket, OLDB, etc ...)
        
        """
        if data is None:
            return self.fget()
        return data[self]
        
    def set(self, value, data=None):
        """ Set node value 
        
        If the optional data dictionary is given `data[self] = value`, otherwise  self.fset(value) is used 
        """
        value = self.parse(value)
                
        if data is None:
            self.fset(value)
        else:
            data[self] = value
    
    ### ############################################
    #
    # To be implemented by the inerated class  
    #
    
    def fget(self):
        """ This is the function we need to implement to get real data """
        raise NotImplementedError('fget')
    
    def fset(self, value):
        """ This is the function we need to implement to set real data """
        raise NotImplementedError('fset')
                


def nodeproperty(name,  *args, **kwargs):
    """ A decorator for a quick node creation 
    
    This shall be implemented in a parent interface or any class 
    
    Args:
        name (str) : name of the node. The key of the node will be parent_key.name
        \*args, \*\*kwargs: All other arguments necessary for the node construction
    """
    return BaseNode.prop(name, *args, **kwargs)

def node(key: Optional[str] =None):
    """ This is a node decorator 
    
    This allow to quickly embed any function in a node without having to subclass Node
    
    The build node will be readonly, for a more complex behavior please subclass a BaseNode
    
    Args:
        key (str): string key of the node
    
    Returns:
       func_setter: a decorator for the fget method  
       
    Example:
    
    A simulator of value:
    
    ::
        
        @node('temperature')
        def temperature():
            return np.random.random()*3 + 20
        
        temperature.get()
        
    Node returning the local float time in seconds 
    
    :: 
        
        @node('local_time')
        def local_time():
            return time.time()
    """
    node = BaseNode(key)
    def set_fget(func):
        node.fget = func
        if hasattr(func, "__func__"):
            node.__doc__= func.__func__.__doc__
        else:
            node.__doc__ = func.__doc__
        return node
    return set_fget


def setitem(obj,k,v):
    obj[k] = v

class NodesReader:
    def __init__(self, nodes=tuple()):
        self._input_nodes = nodes # need to save to remenber the order
        self._dispatch, self._aliases = {}, []
        for node in nodes:
            self.add(node) 
            
    def add(self, node):
        # if no _sid, this ia an alias or a standalone node and should be call 
        # at the end
        
        # None object are ignored 
        if node is None:
            return 
                
        if isinstance(node, (tuple, list, set)):
            for n in node:
                self.add(n)
         
        sid = getattr(node, 'sid', None)
        if sid is None:
            self._aliases.append( node )
            for n in getattr(node, "nodes", []):            
                self.add(n)
            return         
        try:
            collection = self._dispatch[sid]
        except KeyError:
            collection = node.read_collector()
            self._dispatch[sid] = collection
        collection.add(node)
    
    def clear(self):
        self._dispatch.clear()
        self._aliases.clear()
    
    def read(self, data=None):
        """ read all node values 
        
        nodes are first grouped by sid and are red in one call is the server allows it 
        """
        # starts with the UA nodes 
        
        # :TODO: At some point, this should be asynchrone 
        for sid, collection in self._dispatch.items(): 
            collection.read(data)       
                
        # aliases are treated at the end, data should have all necessary real nodes for 
        # the alias 
        # We need to start from the last as Aliases at with lower index can depend 
        # of aliases with higher index
        aliases = self._aliases
        flags = [False]*len(aliases)
        for i, alias in reversed(list(enumerate(aliases))):
            if not flags[i]: 
                data[alias] = alias.get(data)                       
                flags[i]= True

class NodesWriter:
    def __init__(self, node_values):
        
        self._dispatch = {}
        
        # start with aliases, returned values are set inside 
        # the node_values dictionary
        for node, value in node_values.items():
            sid = getattr(node, 'sid', None)
            if sid is None:
                node.set(value, node_values)
        
        for node, value in node_values.items():
            sid = getattr(node, 'sid', None)
            if sid is not None:
                self.add(node, value)
    
    def clear(self):
        self._dispatch.clear()
        self._node_by_key.clear()
        
    def add(self, node, value):
        try:
            collection = self._dispatch[node.sid]
            
        except KeyError:
            collection = node.write_collector()
            self._dispatch[node.sid] = collection
        
        collection.add(node, value)
        
    def write(self) -> None:                
        # :TODO: At some point, this should be asynchrone 
        for collection in  self._dispatch.values():
            collection.write()


def new_node(type_, *args, **kwargs):
    """ Create a new node for a given type 
    
    ::
        new_node( tpe, *args, **kwargs) 

    Is just a short cut for

    ::
        get_node_class(tpe)(*args, **kwargs)
    
    """
    cls = get_node_class(type_)
    return cls(*args, **kwargs) 
        
        
def to_node_class(_func_: Callable =None, *, type: Optional[str] = None):
    if _func_ is None:
        def node_func_decorator(func):
            return _node_func(func, type)
        return node_func_decorator
    else:
        return _node_func(_func_, type)

def _node_func(func, type_):
    if not hasattr(func, "__call__"):
        raise ValueError(f"{func} is not callable")
    
    try:    
        s = signature(func)
    except ValueError: # assume it is a builtin class with one argument 
        conf_args = {}
        obj_args = []
    else:
        
        conf_args = {}
        obj_args = []
        for a,p in s.parameters.items():
            if p.default == _empty:
                if a not in ['com', 'localdata', 'key']:
                    raise ValueError("All arguments must be keyword or one of 'com', 'localdata' or 'key'")
                obj_args.append(a)
            else:    
                if p.annotation == _empty:
                    conf_args[a] = p.default 
                else:
                    conf_args[a] = (p.annotation,p.default)
                    
    extras = {}
    if type_ is None:        
        if  "type" in conf_args:
            type_  = conf_args['type']            
        else:
            type_ = func.__name__
            extras['type'] = type_
    else:
        extras['type'] = type_
                
    Config = create_model(type_+"Config", **extras, **conf_args, __base__= BaseNode.Config)
        
    if conf_args or obj_args: 
        conf_args_set = set(conf_args)        
                
        if obj_args:        
            def fget_method(self):
                c = self.config
                return func(*[getattr(self,__a__) for __a__ in obj_args], **{a:getattr(c,a) for a in conf_args_set})
        else:
            # def fget_method(self):
            #     return func(**self.config.dict(include=conf_args_set))  # this is slower       
            def fget_method(self):
                c = self.config
                return func(**{a:getattr(c,a) for a in conf_args_set})   
    else:
        def fget_method(self):
            return func()
    try:
        doc = func.__doc__
    except AttributeError:
        doc = None        
    return type(type_+"Node", (BaseNode,), {'Config':Config, 'fget': fget_method, '__doc__':doc})
    



        



from .base import BaseUiLinker, WidgetControl, WidgetFactory, get_widget_factory
from pydantic import BaseModel, validator
from typing import Iterable, Union, Optional, List, Dict
from .io import find_ui
from pydevmgr_core import _BaseObject, BaseData, BaseDevice



from PyQt5.QtWidgets import QLayout, QBoxLayout, QGridLayout, QVBoxLayout, QWidget, QFrame
from PyQt5 import QtCore
from PyQt5.uic import loadUi
import glob



class ConfigViewSetup(BaseModel):
    """ configuration of one view setup 

    Args:
        layout (str) : name of the layout in the ui used to publish widget (default is 'ly_devices')
        dev_type (iterable, str): Accepted device type names, e.g. ["Motor", "Drot"]
        alt_dev_type (str, None) : If not found look at a widget defined for a `alt_dev_type`
        exclude_device (str, iterable): exclude the device with the given names
        exclude_dev_type (str, iterable): exclude device with the given types
        widget_kind (str): string defining the widget kind (default, 'ctrl')
        alt_layout (iterable, str): Alternative layout name if `layout` is not found
        
        column, row, columnSpan, rowSpan, stretch, alignment : for layout placement. 
                The use depend of the nature of the layout (Grid, HBox, VBox) 
    
    """
    layout: str = "ly_devices" 
    device: Union[str, Iterable] = "*"
    dev_type: Union[str, Iterable]  = "*"
    alt_dev_type: Optional[List[str]] = None

    exclude_device: Union[str, Iterable] = ""
    exclude_dev_type: Union[str, Iterable]  = ""
    
    widget_kind: str = "ctrl"
    alt_layout: Union[str,Iterable] = [] 
    column: int = 0
    row: int = 0
    columnSpan: int = 1
    rowSpan: int = 1
    
    stretch: int = 0
    alignment: int = 0

    class Config:
        extra = 'forbid'
    
    @validator("alt_layout")
    def _liste_me(cls, value):
        return [value] if isinstance(value, str) else value



class ConfigView(BaseUiLinker.Config):
    """ One item that define a layout compose from a ui resource file and a setup 

    The setup is a list of ViewSetup defining one ui layout and what shall be inside. 
    """
    
    alt_dev_type: Optional[List[str]] = None
    setup: List[ConfigViewSetup] = ConfigViewSetup()
    size: Optional[List] = None
    ui_file: Optional[str] = None

    @validator("ui_file")
    def _validate_ui_file(cls, ui_file):
        """ Check if ui_file exists in resources """
        if ui_file is not None:
            try:
                find_ui(ui_file) 
            except IOError as e:
                raise ValueError(e)
        return ui_file
    @validator('setup', always=True)
    def _validate_setup(cls, setup, values):
        alt_dev_type = values.get('alt_dev_type', None)
        for cv in setup:
            if cv.alt_dev_type is None:
                cv.alt_dev_type = alt_dev_type
        return setup


line_setup =  ConfigView(setup = [ConfigViewSetup(widget_kind="line", device="*")])
ctrl_setup =  ConfigView(setup = [ConfigViewSetup(widget_kind="ctrl", device="*")])


class ConfigDevicesWidget(BaseUiLinker.Config):
    """ Configuration for for several views """
    alt_dev_type: Optional[List[str]] = None
    views : Dict[str,ConfigView] = {"line":line_setup, "ctrl":ctrl_setup}
    
    @validator('views', always=True)
    def _validate_views(cls, views, values):
        alt_dev_type = values.get('alt_dev_type', None)
        for cv in views.values():
            if cv.alt_dev_type is None:
                cv.alt_dev_type = alt_dev_type
            for s in cv.setup:
                if s.alt_dev_type is None:
                    s.alt_dev_type = alt_dev_type
        return views



def insert_widget(
     device: BaseDevice, 
     layout: QLayout, 
     config: ConfigViewSetup = ConfigViewSetup(), *,
     
     default_factory: Optional[WidgetFactory] = None,
        ) -> BaseUiLinker:
    """ Insert one device widget inside a QT Layout object 
    
    Args:
        device (BaseDevice): list of devices 
        layout: (QLayout)
        config:  ConfigViewSetup   
    
    Returns:
       linker (BaseUiLinker): A device linker object (not yet connected to device)
       
    """
    factory = get_widget_factory(config.widget_kind, device.config.type,
                                default=default_factory,
                                alt_dev_type=config.alt_dev_type
                                )       
    linker = factory.build()
    
    widget = linker.widget 
    if isinstance(layout, QBoxLayout): 
        layout.addWidget(widget, config.stretch, QtCore.Qt.AlignmentFlag(config.alignment))
    elif isinstance(layout, QGridLayout):
        layout.addWidget(widget, config.row, config.column, config.rowSpan, config.columnSpan)
    else:
        layout.addWidget(widget)  
    return linker 



def _obj_to_match_func(obj):
    if not obj:
        return lambda name: False 
    if isinstance(obj, str):
        return lambda name: glob.fnmatch.fnmatch(name, obj)
    elif hasattr(obj, "__iter__"): 
        return  lambda name: name in obj

def filter_devices(devices, config: ConfigViewSetup = ConfigViewSetup() ):
        c = config
        output_devices = []
        match_device = _obj_to_match_func(c.device)
        match_type   = _obj_to_match_func(c.dev_type)
        
        exclude_match_device = _obj_to_match_func(c.exclude_device)
        exclude_match_type   = _obj_to_match_func(c.exclude_dev_type)
        for device in devices:        
            if exclude_match_device(device.name): continue
            if exclude_match_type(device.config.type): continue
            if match_device(device.name) and match_type(device.config.type):
                output_devices.append(device)  
        return output_devices


def find_layout(ui: QWidget, config:  ConfigViewSetup = ConfigViewSetup() ):
        """ find a layout from a parent ui according to config 
        
        Look for a layout named as .layout properties. If not found look inside 
        the .alt_layout list property. 
        """
        layout = ui.findChild(QLayout, config.layout)
        if layout is None:
            for ly_name in config.alt_layout:
                layout = ui.findChild(QLayout, ly_name)
                if layout: break
            else:
                raise ValueError(f"Cannot find layout with the name {layout!r} or any alternatives")
        return layout
   



class ViewUiLinker(BaseUiLinker):
    """ A widget linker for one single view """
    Config = ConfigView
    _device_linkers  = None       
    
    class Widget(QFrame):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            # make a very simple widget with one single layout named ly_devices
            # do not change the layouut_name 
            #layout = QVBoxLayout(objectName="ly_devices") # give a default name for the layout 
            #self.setLayout(layout)
            loadUi(find_ui('simple_devices_frame.ui'), self)

    @classmethod
    def new_widget(self, config):
        if config.ui_file:
            return loadUi(find_ui(config.ui_file))
        else:
            return self.Widget()

    def _link(self, downloader):

        for device, linker in self._device_linkers:
            linker.connect( downloader, device )
        


    def connect(self, downloader, obj_list, data = None):
        if data is None:
            data = self.new_data()
        self.setup_ui( obj_list , data)
        self._link(downloader)
        
        return WidgetControl( self, downloader, obj_list, data )
       
    def disconnect(self):
        super().disconnect()
        if self._device_linkers:
            for _, linker in self._device_linkers:
                linker.disconnect()
    
    def setup_ui(self, obj_list: List[_BaseObject], data: BaseData):
        self.disconnect()
        self.clear()
        
        # pairs of device/linker is recorded they are necessary to connect/disconnect etc ... 
        self._device_linkers = []
        for setup in self.config.setup:
            layout= find_layout( self.widget, setup )
            
            for device in filter_devices(obj_list, setup):
                linker = insert_widget(device, layout, setup)
                self._device_linkers.append( (device, linker ))

    def clear(self):
        if self._device_linkers:
            for d,l in self._device_linkers:
                l.widget.setParent(None)
    
class DevicesWidgetLinker(BaseUiLinker):
    """ A widget linker containing several defined views """
    Config = ConfigDevicesWidget
    
    class Data(BaseModel):
        current_view: str = "line"
    
    class Widget(QWidget):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.container_layout = QVBoxLayout()
            self.setLayout(self.container_layout)
    
    _view_linker = None

    def _link(self, downloader):
        self._view_linker._link(downloader)

    def connect(self, downloader, obj_list, data = None):
        if data is None:
            data = self.new_data()
        self.setup_ui(obj_list, data)
        #self._view_linker.connect(downloader, obj_list)
        self._link(downloader)

        return WidgetControl( self, downloader, obj_list, data )

    def disconnect(self):
        super().disconnect()
        if self._view_linker:
            self._view_linker.disconnect()
    
    def clear(self):
        if self._view_linker:
            self._view_linker.clear()
            self._view_linker.widget.setParent(None)
        
    def setup_ui(self, obj_list: List[_BaseObject], data: BaseData):
        self.disconnect()
        self.clear()
        self._view_linker = None

        view = self.config.views[data.current_view]

        vl = ViewUiLinker(config=view)
        self.widget.container_layout.addWidget(vl.widget)
        vl.setup_ui( obj_list,  vl.new_data() ) 
        self._view_linker = vl



from ._class_recorder import get_rpc_class, KINDS
from ._core_base import (_BaseObject, _BaseProperty)
                           
from ._core_com import BaseCom

from .. import io
from typing import Dict, List, Callable, Union , Optional, Any, Type
from pydantic import create_model, BaseModel
from ._core_parser import parser, AnyParserConfig
from ._core_obj_dict import ObjDict
from inspect import signature , _empty

import weakref

class BaseRpcConfig(_BaseObject.Config):
    kind: KINDS = KINDS.RPC.value
    type: str = ""
    
    arg_parsers: Optional[List[Union[AnyParserConfig, List[Union[str, Callable]], str, Callable]]] = []
    kwarg_parsers: Optional[Dict[str,Union[AnyParserConfig, List[Union[str, Callable]], str, Callable]]] = {}
   
    @classmethod
    def validate_kind(cls, kind):
        if kind:
            if kind!=KINDS.RPC:
                raise ValueError(f'expecting a Rpc kind, got a {kind!r}')
        return kind    
    


class BaseCallCollector:
    """ The Read Collector shall collect all nodes having the same sid and read them in one call
    
    - __init__ : should not take any argument 
    - add : take one argument, the Node. Should add node in the read queue 
    - read : takes a dictionary as arguement, read the nodes and feed the data according to node keys 
    
    The BaseReadCollector is just a dummy implementation where nodes are red one after the other     
    """
    def __init__(self):
        self._rpcs = []
    
    def add(self, rpc, args, kwargs):
        self._rpcs.append((rpc, args, kwargs))
        
    def call(self):        
        for rpc, args, kwargs in self._rpcs:
            rpc.rcall(*args, **kwargs)
                
class RpcError(RuntimeError):
    """ Raised when an rpc method is returning somethingelse than 0

        See rcall method of RpcNode
    """
    rpc_error = 0

class RpcProperty(_BaseProperty):    
    fcall = None
    
    def caller(self, func):
        """ decoraotr to define the fcall function """
        self.fcall = func
        return self # must return self
    
 
    def _finalise(self, parent, rpc):
        if self.fcall:
            parent_wr = weakref.ref(parent)
            def fcall(*args, **kwargs):
                return self.fcall(parent_wr(), *args, **kwargs)
            rpc.fcall = fcall
    
    def __call__(self, func):
        """ The call is used has fcall decorator """
        self.fcall = func
        return self


class RpcDict(ObjDict):
    class Config(ObjDict.Config):
        kinds = set([KINDS.RPC])
        default_kind = KINDS.RPC

class BaseRpc(_BaseObject):
    
    Config = BaseRpcConfig
    Property = RpcProperty
    Dict = RpcDict
    
    _arg_parsers = None
    _kwarg_parsers = None
    def __init__(self, 
           key: Optional[str] = None, 
           config: Optional[Config] =None, 
           **kwargs
        ) -> None:  
        super().__init__(key, config=config, **kwargs)
        
        if self.config.arg_parsers:
            _arg_parsers = []
            for i,p in enumerate(self.config.arg_parsers):                
                _arg_parsers.append(parser(p)) 
                self.config.arg_parsers[i] = _arg_parsers[i].config
            self._arg_parsers = _arg_parsers
        if self.config.kwarg_parsers:
            _kwarg_parsers = {}
            for k,p in self.config.kwarg_parsers.items():
                _kwarg_parsers[k] = parser(p)
                self.config.kwarg_parsers[k] = _kwarg_parsers[k].config    
            self._kwarg_parsers  = _kwarg_parsers


    @property
    def sid(self):
        """ default id server is 0 
        
        The sid property shall be adujsted is the CallCollector
        """
        return 0
    
    
    def get_error_txt(self, rpc_error):
        """ Return Error text from an rpc_error code """
        return "Not Registered Error"
    
    def call_collector(self):
        """ Return a collector for method call """
        return BaseCallCollector()
    
    def parse_args(self, args, kwargs):
        """ Modify in place a list of args and kwargs thans to the defined arg_parsers and kwarg_parsers """
        if self._arg_parsers:
            for i,(p,a) in enumerate(zip(self._arg_parsers, args)):
                args[i] = p(a) 
        if self._kwarg_parsers:
            for k,a in kwargs.items():
                try:
                    p = self._kwarg_parsers[k]
                except KeyError:
                    pass
                else:
                    kwargs[k] = p(a)
                
    def call(self, *args, **kwargs):
        """ Call the method and return what return the server 
        
        this will mostly return an integer which shall be 0 if success
        
        .. seealso::
        
           :func:`BaseRpc.rcall` method
          
        """
        args = list(args)
        self.parse_args(args, kwargs)
        return self.fcall(*args, **kwargs)
    
    def rcall(self, *args, **kwargs):
        """ Call the Rpc Method but raised an exception in case of an error code is returned """
        e = self.get_error(self.call(*args, **kwargs))
        if e:
            raise e
    
    def get_error(self, rpc_return):
        if rpc_return:
            e = RpcError("RPC ({}): {}".format(rpc_return, self.get_error_txt(rpc_return)))
            e.rpc_error = rpc_return
            return e
    
    def fcall(self, *args, **kwargs):
        raise NotImplementedError('fcall')



class ComRpc(BaseRpc):
    """ a BaseRpc with a com attribute """
    class Config(BaseRpc.Config):
        type: str= "ComRpc"
        com: Optional[BaseCom.Config] = None
    
    def __init__(self, 
           key: Optional[str], 
           config: Optional[Config] =None, *, 
           com: Optional[Any] =None,  
           **kwargs
        ) -> None:
        self._com = com         
        super().__init__(key, config=config, **kwargs)

    @property
    def com(self):
        return self._com
    
    @classmethod
    def new_args(cls, parent, config):
        d = super().new_args(parent, config)
        d.upadte(com=parent.com)
        return d        
    
        

def rpcproperty(name, *args, **kwargs):
    """ A decorator for a quick rpc creation 
    
    This shall be implemented in a parent interface or any class 
    
    Args:
        cls (class, optional): default is :class:`BaseRpc` used to build the rpc  
        **kwargs: All other arguments necessary for the node construction
    """
    return BaseRpc.prop(name, *args, **kwargs)


def to_rpc_class(_func_: Callable = None, *, type: Optional[str] = None) -> Type[BaseRpc]:
    """ Create a Rpc Class from a function 
    
    This is a conveniant function to quickly create a Rpc Class. 
    As this is a lazy and durty implementation they are some naming convention to respect for the 
    function arguments. 
    The function can have the signature : 
    
    - `f(arg1, arg2, ..., key1=val1, key2=val2, ...)`
    
    Eventualy before the args 1 to 3 positional arguments can be named key, com and localdata 
    
    - `f(key, arg1, arg2, ..., key1=val1, key2=val2, ...)`
    - `f(key, com, arg1, arg2, ..., key1=val1, key2=val2, ...)`
    - `f(com, localdata, key, arg1, arg2, ..., key1=val1, key2=val2, ...)`
    - etc ...
    
    At call, com, localdata and key will be replaced by the corresponding Rpc instance attribute other
    positional argument are parsed to the function. 
    The keyowrd become parameters of the Config class 
    
    The to_rpc_class can be used as decorator or function 
    
    Example:
    
        >>> from pydevmgr_core import to_rpc_class
        >>> @to_rpc_class
        >>> def Echo(key, value, end: str = "\n"):
        >>>     print( key, "is called with value", value,  end=end)
      
        >>> Echo.Config()    
        EchoConfig(kind=<KINDS.RPC: 'Rpc'>, type='Echo', arg_parsers=None, kwarg_parsers=None, arg_parsers_config=[], kwarg_parsers_config={}, com=None, end='\n')
        
        >>> echo = Echo('echo!!', end="||\n")
        >>> echo.call(100)
        echo!! is called with value 100||
        >>> echo.config.end = "*********\n"
        >>> echo.call(100)
        echo!! is called with value 100*********
    
    The above class creation is equivalent of doing
    
    ::
    
       class Echo(BaseRpc):
            class Config(BaseRpc.Config):
                end: str = "\n"
            def fcall(self, value):
                print( self.key, "is called with value", value,  end=self.config.end)
    
    The penalty of the durty class creation is not so big: 1.5 to 2 times slower for the call 
    process only. Probably that what is inside the function will dominate the execution time anyway  
    """
    if _func_ is None:
        def rpc_func_decorator(func):
            return _rpc_func(func, type)
        return rpc_func_decorator
    else:
        return _rpc_func(_func_, type)
    
def _rpc_func(func, type_):
    if not hasattr(func, "__call__"):
        raise ValueError(f"{func} is not callable")
    
    try:    
        s = signature(func)
    except ValueError: # assume it is a builtin class with one argument 
        conf_args = {}
        obj_args = []
        
    else:
        
        conf_args = {}
        obj_args = []
        poasarg = False
        for a,p in s.parameters.items():
            if p.default == _empty:
                if a in ['com', 'localdata', 'key']:
                    if poasarg:
                        raise ValueError("Pos arguments must be after or one of 'com', 'localdata' or 'key'")
                    obj_args.append(a)
                else:
                    poasarg = True        
            else:                
                if p.annotation == _empty:
                    conf_args[a] = p.default 
                else:
                    conf_args[a] = (p.annotation,p.default)
                    
                    
    extras = {}
    if type_ is None:        
        if  "type" in conf_args:
            type_  = conf_args['type']            
        else:
            type_ = func.__name__
            extras['type'] = type_
    else:
        extras['type'] = type_
        
    Config = create_model(type_+"Config", **extras, **conf_args, __base__=BaseRpc.Config)
        
    if conf_args or obj_args: 
        conf_args_set = set(conf_args)        
        
        if obj_args:        
            def fcall_method(self, *args):
                c = self.config
                return func(*[getattr(self,__a__) for __a__ in obj_args], *args, **{a:getattr(c,a) for a in conf_args_set})
        else:
            def fcall_method(self, *args):   
                c = self.config         
                return func(*args, **{a:getattr(c,a) for a in conf_args_set})        
            
    else:
        def fcall_method(self, *args):
            return func(*args)
    try:
        doc = func.__doc__
    except AttributeError:
        doc = None           
    return type(type_+"Rpc", (BaseRpc,), {'Config':Config, 'fcall': fcall_method, '__doc__':doc})
    




        



""" some nodes that required numpy to be executed if numpy is not installed theses nodes will not be loaded """

from ._core_objects import  BaseNode, NodeAlias, NodeAlias1, record_class
from collections import deque
from typing import Optional, List, Tuple
import numpy as np
from enum import Enum 

__all__ = [
"NoiseNode",
"NoiseAdderNode",
"HistogramNode",
"MeanFilterNode", 
"MaxFilterNode", 
"MinFilterNode", 
"VarianceFilterNode", 
"RmsFilterNode", 
"MedianFilterNode", 
"PickToValleyFilterNode"
]

def random(mean, scale, size):
    return np.random.random(size)*scale+mean


class RANDOM_FUNC(Enum):
    RANDOM = "random"
    NORMAL = "normal"
    LOGNORMAL = "lognormal"
RANDOM_FUNC.RANDOM.func = random
RANDOM_FUNC.NORMAL.func = np.random.normal
RANDOM_FUNC.LOGNORMAL.func = np.random.lognormal

@record_class
class NoiseNode(BaseNode):
    class Config(BaseNode.Config):
        type: str = "Noise"
        mean: float = 0.0 # Mean ("centre") of the distribution
        scale: float = 1.0 # Standard deviation (spread or "width") of the distribution
        size: Optional[List[int]] = None
        distribution: RANDOM_FUNC = RANDOM_FUNC.NORMAL
    
    def fget(self):
        c = self.config
        return RANDOM_FUNC(c.distribution).func(c.mean, c.scale, c.size)
        
@record_class
class NoiseAdderNode(NodeAlias1):
    class Config(NodeAlias1.Config):
        type: str = "NoiseAdder"
        scale: float = 1.0 # Standard deviation (spread or "width") of the distribution
        distribution: RANDOM_FUNC = RANDOM_FUNC.NORMAL
    
    def fget(self, value):
        c = self.config
        return RANDOM_FUNC(c.distribution).func(value, c.scale)
        
@record_class
class HistogramNode(NodeAlias1):
    class Config(NodeAlias1.Config):
        type: str = "Histogram"
        bins: Tuple[float,float,int] = (-100,100,100)        
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.reset()
    def reset(self):
        min,max,n = self.config.bins
        self._bins = np.linspace(min,max,n)
        self._hist = [0]*(len(self._bins)-1)
        
    def fget(self, value):
        min,max,_ = self.config.bins
        if value<min or value>max:
            return self._hist
        i = np.digitize(value, self._bins)    
        self._hist[i-1] += 1
        return self._hist
            

class _Filter(NodeAlias1):
    class Config(NodeAlias1.Config):
        nval : int = 10
        
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._data = deque([], self.config.nval)
    
    def fget(self, value):
        self._data.append(value)
        return self._func(self._data)

@record_class
class MeanFilterNode(_Filter):
    class Config(_Filter.Config):
        type = "MeanFilter"    
    _func = staticmethod(np.mean)    

@record_class
class MaxFilterNode(_Filter):
    class Config(_Filter.Config):
        type = "MaxFilter"    
    _func = staticmethod(np.max)  
      
@record_class
class MinFilterNode(_Filter):
    class Config(_Filter.Config):
        type = "MinFilter"    
    _func = staticmethod(np.min)

@record_class
class VarianceFilterNode(_Filter):
    class Config(_Filter.Config):
        type = "VarianceFilter"
    @staticmethod
    def _func(data):
        data = np.asarray(data)
        m = np.mean(data)
        return (data-m)**2/len(data)

@record_class
class RmsFilterNode(_Filter):
    class Config(_Filter.Config):
        type = "RmsFilter"
    @staticmethod
    def _func(data):
        data = np.asarray(data)
        m = np.mean(data)
        return np.sqrt(  (data-m)**2/len(data) )         

@record_class
class MedianFilterNode(_Filter):
    class Config(_Filter.Config):
        type = "MedianFilter"    
    _func = staticmethod(np.median)

@record_class
class PickToValleyFilterNode(_Filter):
    class Config(_Filter.Config):
        type = "PickToValleyFilter"
    @staticmethod
    def _func(data):
        return  np.max(data)-np.min(data)  
   
        



from ._core_node import BaseNode, BaseNode
from ._class_recorder import record_class
from typing import Any


class LocalNodeConfig(BaseNode.Config):
    default: Any = None
    type = "Local"

class StaticNodeConfig(BaseNode.Config):    
    type = "Static"
    value: Any
    
@record_class
class StaticNode(BaseNode):
    """ """
    Config = StaticNodeConfig    
    def fget(self):        
        return self.config.value    
    
@record_class
class LocalNode(BaseNode):
    """ A Node getting and setting values inside the localdata dictionary 
    
    The node can only work if its localdata dictionary is an attribute
    """    
    Config = LocalNodeConfig
    def fget(self):
        if self.localdata is None:
            return self.config.default                
        return self.localdata.get(self.key,self.config.default)
        
    def fset(self, value):
        self.localdata[self.key] = value
        
        



from ._core_objects import NodesWriter, BaseNode
from .download import BaseDataLink

from typing import Union, Optional, Callable, Any, Dict 
import time 


class Uploader:
    """ An uploader object to upload data to the PLC 
    
    The values to upload is defined in a dictionary of node/value pairs. 
    
    Not sure their is a strong use case for this. Maybe if pydevmgr is used as server instead of client 
    
    Args:
        node_dict_or_datalink (dict, :class:`DataLink`):
             Dictionary of node/value pairs like ``{ motor.cfg.velocity : 4.3 }``
             Or a :class:`pydevmgr_core.DataLink` object.  
        callback (callable, optional): callback function after each upload
    
    Example:
        
    ::
    
        >>> values  = {mgr.motor1.velocity: 1.0, mgr.motor2.velocity: 2.0}
        >>> uploader = Uploader(values)
        >>> t = Thread(target=uploader.runner)
        >>> t.start()
        >>> uploader[mgr.motor1.velocity] = 1.4 # will be uploaded at next trhead cycle 
    
    ::
    
        from pydevmgr_elt import DataLink, NodeVar
        from pydantic import BaseModel 
        
        class Config(BaseModel):
            backlash: NodeVar[float] = 0.0
            disable: NodeVar[bool] = False
        
        >>> conf = Config()
        >>> Uploader( DataLink(mgr.motor1.cfg, conf) ).upload()
            
    .. seealso::
    
       :func:`upload`:  equivalent to Uploader(node_values).upload() 
       
       
    """
    def __init__(self, 
          node_dict_or_datalink: Union[Dict[BaseNode,Any], BaseDataLink], 
          callback: Optional[Callable] = None
        ) -> None:
        
        if node_dict_or_datalink is None:
            node_values = {}
            datalink = None
        elif isinstance(node_dict_or_datalink, BaseDataLink):
            datalink = node_dict_or_datalink
            node_values = {}
            datalink._upload_to(node_values)
        else:
            node_values = node_dict_or_datalink
            datalink = None    
            
        
        self.node_values = node_values 
        self.datalink = datalink
        self.callback = callback
    
    def __has__(self, node):
        return node in self._node_values
        
    def upload(self) -> None:
        """ upload the linked node/value dictionaries """
        if self.datalink:
            self.datalink._upload_to(self.node_values)
        
        NodesWriter(self.node_values).write() 
        if self.callback:
            self.callback()
                  
    def run(self, 
          period: float = 1.0, 
          stopsignal: Callable = lambda : False, 
          sleepfunc:  Callable = time.sleep
        ) -> None:
        """ Run the upload infinitly or until stopsignal is True 
        
        Args:
            period (float, optional): period of each upload cycle
            stopsignal (callable, optional): A function called at each cycle, if True the loop is break
                       and run returns    
        """
        while not stopsignal():
            s_time = time.time()
            self.upload()
            sleepfunc( max( period-(time.time()-s_time), 0))
    
    def runner(self, 
          period: float = 1.0, 
          stopsignal: Callable = lambda : False, 
          sleepfunc:  Callable = time.sleep
        ) -> Callable:
        """ return a function to updload 
        
        this is designed in particular to be used in a target Thread
        
        Args:
            period (float, optional): period of each upload cycle
            stopsignal (callable, optional): A function called at each cycle, if True the loop is break
                       and run returns
        
        Example:
            
            ::
            
                >>> values  = {mgr.motor1.velocity: 1.0, mgr.motor2.velocity: 2.0}
                >>> uploader = Uploader(values)
                >>> t = Thread(target=uploader.runner)
                >>> t.start()
                >>> values[mgr.motor1.velocity] = 1.2 # will be updated at next thread cycle
                               
        """           
        def run_func():
            self.run( period=period, sleepfunc=sleepfunc, stopsignal=stopsignal)
        return run_func
    


def upload(node_dict_or_datalink : Union[Dict[BaseNode,Any], BaseDataLink] ) -> None:
    """ write node values to the remotes
    
    Args:
        node_dict_or_datalink (dict):
             Dictionary of node/value pairs like  ``{ motor.cfg.velocity : 4.3 }``
             Or a :class:`pydevmgr_core.DataLink` object.  
                
    .. note:: 
        
        The input dictionary has pairs of node/value and not node.key/value      
    """
    NodesWriter(node_dict_or_datalink).write()    



import yaml
import os
import re
from pydantic import BaseModel
from typing import Tuple, Optional, List, Dict, Optional, Callable, Any
from ._yaml_loader import PydevmgrLoader

class IOConfig(BaseModel):
    cfgpath : str = 'CFGPATH'
    yaml_ext : Optional[List[str]] = ('.yml', '.yaml')
    YamlLoader = PydevmgrLoader
    
ioconfig = IOConfig()




_re_path_pattern = re.compile( '^([^\\[]+)\\[([^\\]]*)\\]$' )
def parse_file_name(file_name: str):
    """ split a file name into real file and path tuple"""
    g = _re_path_pattern.search(file_name)
    if not g:
        return file_name, None
    else:
        file, path = g[1], g[2]
         
        return file.strip(' '), tuple( p for p in path.strip(' ').split('.') if p)


def load_config(file_name: str, ioconfig: IOConfig = ioconfig) -> Dict:
    """ Load a configuration file into a dictionary
    
    The file name is given, it should be inside one of the path defined by the $CFGPATH 
    environment variable (or the one defined inside ioconfig.cfgpath). 
    
    Alternatively the path can be an absolute '/' path 

    For now only yaml configuration files are supported 
    
    At the end of the file name can be a path to a target value inside the configuration. Path are string separated by
    '.'
    
        e.g.
        
        ::
            
            assert load_config( 'myconfig.yml[a.b.c]' ) == load_config('myconfig.yml')['a']['b']['c']

    Args:
        file_name (str): config file name. Should be inside one of the path defined by the $CFGPATH 
        environment variable. Alternatively the file_name can be an abosolute path
        ioconfig (optional): a :class:`IOConfig`        

    Returns:
        cfg (dict): config dictionary
    """
    file_name, path = parse_file_name( file_name)
    return read_config(find_config(file_name, ioconfig = ioconfig), ioconfig = ioconfig, path=path)


def read_config(file: str, ioconfig: IOConfig =ioconfig, path=None) -> Dict:    
    """ Read the given configuration (yaml) file 
    
    Args:
        file (str): real file path, shall be a yaml file. 
    """
    if not path:
        with open(file) as f:
            return yaml.load(f.read(), Loader=ioconfig.YamlLoader)
    else:
        with open(file) as f:
            d = yaml.load(f.read(), Loader=ioconfig.YamlLoader)
            for p in path:
                if p:
                    try:
                        d = d[p]
                    except KeyError:
                        raise ValueError(f"Cannot resolve path {path!r} inside {file!r}")
            return d


def load_yaml(input: str, ioconfig: IOConfig =ioconfig) -> Any:
    return yaml.load(input, Loader=ioconfig.YamlLoader)

def find_config(file_name, ioconfig: IOConfig =ioconfig):
    """ find a config file and return its absolute path 
    
    Args:
        file_name (str): config file name. Should be inside one of the path defined by the $CFGPATH 
        environment variable. Alternatively the file_name can be an abosolute path
        ioconfig (optional): a :class:`IOConfig`        
    """
    path_list = os.environ.get(ioconfig.cfgpath, '.').split(':')
    for directory in path_list[::-1]:
        path = os.path.join(directory, file_name)
        if os.path.exists(path):
            return  path
    raise ValueError('coud not find config file %r in any of %s'%(file_name, path_list))

# overwrite the PydevmgrLoader.find_config method 
PydevmgrLoader.find_config = staticmethod(find_config)

    
def explore_config(filter: Optional[Callable] =None, ioconfig: IOConfig = ioconfig):
    """ Iterator on all config files found inside directories of the $CFGPATH environment variable 
    
    The iterator returns pairs of (relateve_path, root_directory)   

    Args:
        filter (None, callable, optional): if given it will receive the content for each file
               to filter
        ioconfig (optional): a :class:`IOConfig`

    Example:

    ::

        >>> list(explore_config(  lambda d: d['kind']=='Device' ))

    """
        
    path_list = os.environ.get(ioconfig.cfgpath, '.').split(':')
    found = set()
    
    for root in path_list[::-1]:
        for r, d, files in os.walk(root):
            for file in files:
                body, ext = os.path.splitext(file)
                if ext in ioconfig.yaml_ext:
                    p = os.path.relpath( os.path.join(r,file), root )
                    if not p in found:
                        if filter:
                            if filter(read_config(os.path.join(r,file))):
                                yield  p, root
                        else:
                            yield  p, root                                    
                    found.add(p)



def append_cfgpath( path: str, ioconfig: IOConfig = ioconfig):
    """ Add a new path to the $CFGPATH environment variable 

    Args:
        path (str): new config file directory 
        ioconfig (optional): a :class:`IOConfig` 
    """
    envpath = os.environ.get(ioconfig.cfgpath, '').split(':')
    envpath.append(path)
    os.environ[ioconfig.cfgpath] = ":".join(envpath)






from setuptools import setup, find_packages
import sys

# Python 3.0 or later needed
if sys.version_info < (3, 6, 0, 'final', 0):
    raise SystemExit('Python 3.6 or later is required!')



setup(
    name= 'pydevmgr_core',
    version= '0.4.4', # https://www.python.org/dev/peps/pep-0440/
    author='Sylvain Guieu',
    author_email='sylvain.guieu@univ-grenoble-alpes.fr',
    packages=find_packages(), 
    #scripts=scripts,
    #data_files=data_files,
    license='Creative Commons Attribution-Noncommercial-Share Alike license',
    long_description=open('README.md').read(),
    install_requires=['pyyaml',  'pydantic>=1.9',  'py_expression_eval'],
    
    extras_require={
        "QT":  ["pyqt5"],
    },
    
    dependency_links=[],
    long_description_content_type='text/markdown',
    
    include_package_data=True, 
    package_data= {
        'pydevmgr_core':    ["resources/*.yml"], 
        'pydevmgr_core_qt': ["uis/*.ui"]
    }, 
    entry_points = {
        'console_scripts': [],
    }
)



from ._core_objects import NodesReader, BaseNode, kjoin, _BaseObject

import time
from collections import  OrderedDict

from typing import Any, Dict, Iterable, Union, Optional, Callable


class DataView:
    def __init__(self, 
            data: Dict[BaseNode,Any], 
            prefix: Optional[Union[str, _BaseObject]] = None
          ) -> None:
        self._data = data
        if prefix is None:
            prefix = ""
        if not isinstance(prefix, str):
            prefix = prefix.key 
        
        if not prefix:
            key2data = {n.key:n for n in data if hasattr(n, "key") }
        else:                    
            key2data = {}
            pref = prefix+"."
            lp = len(pref)
            for n in data:
                if hasattr(n, "key") and n.key.startswith(pref):
                    key2data[n.key[lp:]] = n
        
        self._key2data = key2data    
    
    def __repr__(self):
        return repr({k:self._data[n] for k,n in self._key2data.items() })
    
    def __str__(self):
        return str({k:self._data[n] for k,n in self._key2data.items() })
        
    def __getitem__(self, item):
        return self._data[self._key2data[item]]
    
    def __setitem__(self, item, value):
        self._data[self._key2data[item]] = value
    
    def __delitem__(self, item):
        del self._data[self._key2data[item]]
    
    def __getattr__(self, attr):
        return self._data[self._key2data[attr]]
    
    def __has__(self, item):
        return item in self._key2data
    
    def update(self, __d__={}, **kwargs) -> None:
        for k,v in dict(__d__, **kwargs).iteritems():
            self._data[self._key2data[k]] = v
    
    def pop(self, item) -> Any:
        """ Pop an item from the root data ! """
        return self._data.pop(self._key2data[item])    
    
    def popitem(self, item) -> Any:
        """ Pop an item from the root data ! """
        return self._data.popitem(self._key2data[item])  
    
    def keys(self) -> Iterable:
        """ D.keys() ->iterable on D's root keys with matching prefix
        
        Shall be avoided to use in a :class:`Prefixed` object
        """
        return self._key2data.__iter__()
    
    def items(self) -> Iterable:
        for k,n in self._key2data.items():
            yield k, self._data[n]
    
    def values(self) -> Iterable:
        for k,n in self._key2data.items():
            yield self._data[n]    
    
    def clear(self) -> None:
        """ D.clear() -> None.  Remove all items from D root with matching prefix
        
        Shall be avoided to use in a :class:`Prefixed` object
        """        
        pref = self._prefix+"."
        for k, n in list(self._key2data.items()):            
            self._data.pop(n)


def _setitem(d,k,v):
    d[k] = v
def _dummy_callback(data):
    pass
def _dummy_trigger():
    return True

class BaseDataLink:
    """ place holder for an instance check """    
    pass


class DownloaderConnection:
    """ Hold a connection to a :class:`Downloader` 
    
    Most likely created by :meth:`Downloader.new_connection` 
    
    Args:
       downloader (:class:`Downloader`) :  parent Downloader instance
       token (Any): Connection token 
    """
    def __init__(self, downloader, token):
        self._downloader = downloader 
        self._token = token 
    
    def _check_connection(self):
        if not self._token:
            raise RuntimeError("DownloaderConnection has been disconnected from its Downloader")
    
    @property
    def data(self) -> dict:
        """ downloader data """
        return self._downloader.data 
           
    def disconnect(self) -> None:
        """ disconnect connection from the downloader 
        
        All nodes related to this connection (and not used by other connection) are removed from the
        the downloader queue. 
        Also all callback associated with this connection will be removed from the downloader 
        
        Note that the disconnected nodes will stay inside the downloader data but will not be updated
        """
        self._downloader.disconnect(self._token)
        self._token = None
        
    def add_node(self, *nodes) -> None:
        """ Register nodes to be downloader associated to this connection 
        
        Args:
            *nodes :  nodes to be added to the download queue
        """ 
        self._check_connection() 
        self._downloader.add_node(self._token, *nodes)
    
    def add_nodes(self, nodes) -> None:
        """ Register nodes to be downloader associated to this connection 
        
        Args:
            nodes :  nodes to be added to the download queue. 
                     If a dictionary of node/value pairs, they are added to the downloader data. 
        """
        self._check_connection() 
        self._downloader.add_nodes(self._token, nodes)
    
    def remove_node(self, *nodes) -> None:
        """ remove  any nodes to the downloader associated to this connection 
        
        Note that the node will stay inside the downloader data but will not be updated 
        
        Args:
            *nodes :  nodes to be removed from the download queue
        """ 
        self._check_connection() 
        self._downloader.remove_node(self._token, *nodes)
    
    def add_datalink(self, *datalinks) -> None:
        """ Register datalinks to the downloader associated to this connection 
        
        Args:
            *datalinks :  :class:`DataLink` to be added to the download queue on the associated downloader
        """
        self._check_connection() 
        self._downloader.add_datalink(self._token, *datalinks)        
    
    def remove_datalink(self, *datalinks) -> None:
        """ Remove any given datalinks to the downloader associated to this connection 
        
        Args:
            *datalinks :  :class:`DataLink` to be removed 
        """
        self._check_connection() 
        self._downloader.remove_datalink(self._token, *datalinks)        
    
    def add_callback(self, *callbacks) -> None:
        """ Register callbacks to be executed after each download of the associated downloader 
        
        Args:            
            *callbacks :  callbacks to be added to the queue of callbacks on the associated downloader       
        """
        self._check_connection() 
        self._downloader.add_callback(self._token, *callbacks)
    
    def remove_callback(self, *callbacks) -> None:
        """ Remove any of given callbacks of the associated downloader 
        
        Args:            
            *callbacks :  callbacks to be remove 
        """
        self._check_connection() 
        self._downloader.remove_callback(self._token, *callbacks)
    
    def add_failure_callback(self, *callbacks) -> None:
        """ Register callbacks to be executed after each download of the associated downloader 
        
        Args:            
            *callbacks :  failure callbacks to be added to the queue of callbacks on the associated downloader       
        """
        self._check_connection() 
        self._downloader.add_failure_callback(self._token, *callbacks)
        
    def remove_failure_callback(self, *callbacks) -> None:
        """ Remove any given callbacks of the associated downloader 
        
        Args:            
            *callbacks :  failure callbacks to be removed 
        """
        self._check_connection() 
        self._downloader.remove_failure_callback(self._token, *callbacks)

class Downloader:
    """ object dedicated to download nodes, feed data and run some callback 

    An application can request nodes to be downloaded and callback to be executed after each 
    success download or each failures. 
    
    Args:    
        nodes_or_datalink (iterable of node or :class:`DataLink`): 
            - An initial, always downloaded, list of nodes 
            - Or a :class:`DataLink` object
        data (dict, optional): A dictionary to store the data. If not given, one is created and
                               accessible through the .data attribute. 
                               This data is made of node/value pairs, the .get_data_view gives however
                               a dictionary like object with string/value pairs. 
                               Each time a new node is added to the downloader it will be added to 
                               the data as ``data[node] = None``. None will be replaced after the 
                               next download.  
                               
        callback (callable, optional): one single function with signature f(), if given always 
                                      called after successful download. 
        trigger (callable, optional): a function taking no argument and should return True or False 
                                      If given the "download" method download nodes only if f() return True. 
                                      Can be used if the download object is running in a thread for instance.
    
    Example: 
    
        A dumy exemple, replace the print_pos by a GUI interface for instance:
        
        ::
                        
            def print_pos(m, data):
                "An application"
                print("Position :",  data[m.stat.pos_actual], data[m.stat.pos_error] )
                
            >>> tins = open_manager('tins/tins.yml')
            >>> tins.connect() 
            >>> downloader = Downloader()
            >>> token = downloader.new_token()
            >>> downloader.add_node(token, tins.motor1.stat.pos_actual, tins.motor1.stat.pos_error )
            >>> downloader.add_callback(token, lambda : print_pos(tins.motor1, downloader.data))
            >>> downloader.download() 
            Position : 3.45 0.003
            >>> downloader.data
            {
            <UaNode key='motor1.pos_error'>: 0.003, 
            <UaNode key='motor1.pos_actual'>:  3.45
            }
            
            >>> downloader.disconnect(token) # disconnect the print_pos function and remove 
                                             # the pos_actual, pos_error node from the list of nodes
                                             # to download (except if an other connection use it)
            
        Same result can be obtained with this exemple: 
        
        :: 
        
            def print_pos(data):
                "An application"
                print("Position :",  data['pos_actual'], data['pos_error'] )
            
            >>> nodes_data = {tins.motor1.stat.pos_actual: -9.99 , tins.motor1.stat.pos_error: -9.99}
            >>> m1_data = DataView(nodes_data, tins.motor1.stat)            
            >>> downloader = Downloader(nodes_data, callback=lambda: print_pos(m1_data))
            >>> downloader.download() 
            Position : 3.45 0.003
            >>> m1_data
            {'pos_error': 0.003, 'pos_actual': 3.45}
        
        
        
                                     
    """
    
    _did_failed = False 
    Connection = DownloaderConnection
    
    def __init__(self,  
            nodes_or_datalink: Union[Iterable, BaseDataLink] = None,  
            data: Optional[Dict] = None, 
            callback: Optional[Callable] = None,
            trigger: Optional[Callable] = None
        ) -> None:
        if data is None:
            data = {}
               
        self._data = data 
        
        if nodes_or_datalink is None:
            nodes = set()
            datalinks = set()
        elif isinstance(nodes_or_datalink, BaseDataLink):
            nodes = set()
            datalinks = set([nodes_or_datalink])
        
        elif isinstance( nodes_or_datalink, dict):
            nodes = set()
            datalinks = set()
            for k,v in nodes_or_datalink.items():
                if isinstance( v, BaseDataLink ):
                    datalinks.add(v)
                else:
                    nodes.add(v)
        
        else:
            nodes = set()
            datalinks = set()
            for v in nodes_or_datalink:
                if isinstance( v, BaseDataLink ):
                    datalinks.add(v)
                else:
                    nodes.add(v)
            



        #nodes = set() if nodes is None else set(nodes)
        if callback is None:
            callbacks = set()
        else:
            callbacks = set([callback])
            
        failure_callbacks = set() 
        
        if callback is None:
            callback = _dummy_callback
        if trigger is None:
            trigger =  _dummy_trigger
        
        self._trigger = trigger
        # Ellipsis is here to define general nodes,datalinks,callbacks, ... independent to connection
        self._dict_nodes = OrderedDict([(Ellipsis,nodes)])
        self._dict_datalinks = OrderedDict([(Ellipsis,datalinks)])
        self._dict_callbacks = OrderedDict([(Ellipsis,callbacks)])
        self._dict_failure_callbacks = OrderedDict([(Ellipsis,failure_callbacks)])
        
        
        self.trigger = trigger
        self._next_token = 1
        
        self._rebuild_nodes()
        self._rebuild_callbacks()
        self._rebuild_failure_callbacks()
    
    def __has__(self, node):
        return node in self._nodes
    
    @property
    def data(self):
        return self._data
    
    def _rebuild_nodes(self):
        nodes = set()
        for nds in self._dict_nodes.values():
            nodes.update(nds)
            for n in nds:
                self._data.setdefault(n,None)
        for dls in self._dict_datalinks.values():
            for dl in dls:
                nodes.update(dl.rnodes)
                for n in dl.rnodes:
                    self._data.setdefault(n,None)
                
        self._nodes = nodes
        self._to_read = NodesReader(nodes)

    def _rebuild_callbacks(self):
        callbacks = set()
        for clbc in self._dict_callbacks.values():
            callbacks.update(clbc)
        self._callbacks = callbacks
    
    def _rebuild_failure_callbacks(self):
        callbacks = set()
        for clbc in self._dict_failure_callbacks.values():
            callbacks.update(clbc)
        self._failure_callbacks = callbacks
    
    
    def new_token(self) -> tuple:
        """ add a new app connection token
        
        Return:
           A token, the token and type itself is not relevant, it is just a unique ID to be used in 
                    add_node, add_callback, add_failure_callback, and disconnect methods 
                    
        .. note::
        
            new_connection() method return object containing a pair of token and downloader and all
                methods necessary to add_nodes, add_callbacks, etc ... 
                
        
        """
        token = id(self), self._next_token
        self._dict_nodes[token] = set()
        self._dict_datalinks[token] = set()
        self._dict_callbacks[token] = set()
        self._dict_failure_callbacks[token] = set()
        
        self._next_token += 1
        # self._rebuild_nodes()
        # self._rebuild_callbacks()
        # self._rebuild_failure_callbacks()
        return token
    
    def new_connection(self):
        """ Return a :class:`DownloaderConnection` object 
        
        The :class:`DownloaderConnection` object contain a token and the downloader in order to have 
        a standalone object to handle the add/remove of queue nodes and callbacks 
        """
        return DownloaderConnection(self, self.new_token())
    
    def disconnect(self, token: tuple) -> None:
        """ Disconnect the iddentified connection 
        
        All the nodes used by the connection (and not by other connected app) will be removed from the 
        download queue of nodes.
        Also all callback associated with this connection will be removed from the downloader 
        
        Note that the discnnected nodes will stay inside the downloader data but will not be updated
         
        Args:
            token : a Token returned by :func:`Downloader.new_token`
        """
        if token is Ellipsis:
            raise ValueError('please provide a real token')
        
        try:
            self._dict_nodes.pop(token)
            self._dict_datalinks.pop(token)
            self._dict_callbacks.pop(token)
            self._dict_failure_callbacks.pop(token)
        except KeyError:
            pass
        
        self._rebuild_nodes()
        self._rebuild_callbacks()
        self._rebuild_failure_callbacks()
    
    def add_node(self, token: tuple, *nodes) -> None:
        """ Register node to be downloaded for an iddentified app
        
        Args:
            token: a Token returned by :func:`Downloader.new_token` 
                   ``add_node(...,node1, node2)`` can also be used, in this case nodes will be added
                   to the main pool of nodes and cannot be removed from the downloader 
            *nodes :  nodes to be added to the download queue, associated to the app
        """   
        self._dict_nodes[token].update(nodes)
        self._rebuild_nodes()
    
    def add_nodes(self, token: tuple, nodes: Union[dict,Iterable]) -> None:
        """ Register nodes to be downloaded for an iddentified app
        
        Args:
            token: a Token returned by :func:`Downloader.new_token` 
                   ``add_node(...,node1, node2)`` can also be used, in this case nodes will be added
                   to the main pool of nodes and cannot be removed from the downloader 
            nodes (Iterable, dict):  nodes to be added to the download queue, associated to the app
                   If a dictionary of node/value pairs, they are added to the downloader data.  
        """
        if isinstance(nodes, dict):
            for node,val in nodes.items():
                self._data[node] = val
        
        self._dict_nodes[token].update(nodes)
        self._rebuild_nodes()
    
    def remove_node(self, token: tuple, *nodes) -> None:
        """ Remove node from the download queue
    
        if the node is not in the queueu nothing is done or raised
        
        Note that the node will stay inside the downloader data but will not be updated 
        
        Args:
            token: a Token returned by :func:`Downloader.new_token`                  
            *nodes :  nodes to be removed 
        """   
        for node in nodes:
            try:
                self._dict_nodes[token].remove(node)
            except KeyError:
                pass 
        self._rebuild_nodes()
    
    def add_datalink(self, token: tuple, *datalinks) -> None:
        """ Register a new datalink
        
        Args:
            token: a Token returned by :func:`Downloader.new_token`
                ``add_datalink(...,dl1, dl2)`` can also be used, in this case they will be added
                to the main pool of datalinks and cannot be remove from the downloader   
            *datalinks :  :class:`DataLink` to be added to the download queue, associated to the token 
        """           
        self._dict_datalinks[token].update(datalinks)
        self._rebuild_nodes()    
    
    def remove_datalink(self, token: tuple, *datalinks) -> None:
        """ Remove a datalink from a established connection
        
        If the datalink is not in the queueu nothing is done or raised
        
        Args:
            token: a Token returned by :func:`Downloader.new_token`
            *datalinks :  :class:`DataLink` objects to be removed         
        """
        for dl in  datalinks:
            try:
                self._dict_datalinks[token].remove(dl)
            except KeyError:
                pass 
        self._rebuild_nodes()
        
    def add_callback(self, token: tuple, *callbacks) -> None:   
        """ Register callbacks to be executed after each download 
        
        The callback must have the signature f(), no arguments.
        
        Args:
            token: a Token returned by :func:`Downloader.new_connection`
            *callbacks :  callbacks to be added to the queue of callbacks, associated to the app
        
        """ 
        self._dict_callbacks[token].update(callbacks)
        self._rebuild_callbacks()
    
    def remove_callback(self, token: tuple, *callbacks) -> None:   
        """ Remove callbacks 
        
        If the callback  is not in the queueu nothing is done or raised
        
        Args:
            token: a Token returned by :func:`Downloader.new_token`
            *callbacks :  callbacks to be removed 
        
        """
        for c in callbacks:
            try:
                self._dict_callbacks[token].remove(c)
            except KeyError:
                pass 
        self._rebuild_callbacks()
    
    
    def add_failure_callback(self, token: tuple, *callbacks) -> None:  
        """ Add one or several callbacks to be executed when a download failed 
        
        When ever occur a failure (Exception during download) ``f(e)`` is called with ``e`` the exception. 
        If a download is successfull after a failure ``f(None)`` is called one time only.
                
        Args:
            token: a Token returned by :func:`Downloader.new_token`
            *callbacks: callbacks to be added to the queue of failure callbacks, associated to the app
        
        """ 
        self._dict_failure_callbacks[token].update(callbacks)
        self._rebuild_failure_callbacks()
    
    def remove_failure_callback(self, token: tuple, *callbacks) -> None:  
        """ remove  one or several failure callbacks 
        
        If the callback  is not in the queue nothing is done or raised
        
        Args:
            token: a Token returned by :func:`Downloader.new_token`
            *callbacks :  callbacks to be removed         
        """ 
        for c in callbacks:
            try:
                self._dict_failure_callbacks[token].remove(c)
            except KeyError:
                pass         
        self._rebuild_failure_callbacks()
    
    
    def run(self, 
            period: float =1.0, 
            stopsignal: Callable =lambda : False, 
            sleepfunc: Callable =time.sleep
        ) -> None:
        """ run indefinitely or when stopsignal return True the download 
        
        Args:
            period (float, optional): period between downloads in second
            stopsignal (callable, optional): a function returning True to stop the loop or False to continue
            
        """
        try:
            while not stopsignal():
                s_time = time.time()
                self.download()
                sleepfunc( max( period-(time.time()-s_time), 0))
        except StopIteration: # any downloader call back can send a StopIteration to stop the runner 
            return 
            
    def runner(self, 
        period: float =1.0, 
        stopsignal: Callable =lambda : False, 
        sleepfunc: Callable =time.sleep
        ) -> Callable: 
        """ Create a function to run the download in a loop 
        
        Usefull to define a Thread for instance
        
        Args:
            period (float, optional): period between downloads in second
            stopsignal (callable, optional): a function returning True to stop the loop or False to continue
        
        Example:
            
            >>> downloader = Downloader([mgr.motor1.substate, mgr.motor1.pos_actual])
            >>> t = Thread( target = downloader.runner(period=0.1) )
            >>> t.start()
            
        """       
        def run_func():
            self.run(period=period, sleepfunc=sleepfunc, stopsignal=stopsignal)
        return run_func
    
    def download(self) -> None:
        """ Execute a download 
        
        Each nodes on the queue are fetched and the .data dictionary is updated
        from new values.
        
        If the Downloader has a trigger method and the trigger return false, nothing is done
        
        """
        
        if not self.trigger(): return 
        
        try:
            self._to_read.read(self._data)
        except Exception as e:
            if self._failure_callbacks:
                self._did_failed = True
                for func in self._failure_callbacks:                    
                    func(e)
            else:
                raise e            
        else:
            # Populate the data links 
            for dls in self._dict_datalinks.values():
                for dl in dls:
                    dl._download_from(self._data)
            
            if self._did_failed:
                self._did_failed = False
                for func in self._failure_callbacks:                    
                    func(None)
                    
            for func in self._callbacks:
                func()
    
    def reset(self) -> None:
        """ All nodes of the downloader with a reset method will be reseted """
        reset(self._dict_nodes)
            
    def get_data_view(self, prefix: str ='') -> DataView:
        """ Return a view of the data in a dictionary where keys are string keys extracted from nodes
        
        If prefix is given the return object will be limited to items with key
        matching the prefix.  
        
        Note: the data view reflect any change made on the rootdata except when new nodes 
        (mathing the prefix) are added. So all necessary nodes shall be added to the downloader 
        before requesting a DataView. 
        
        Args:
           prefix (str, optional): limit the data viewer to a given prefix. prefix can also be an object 
                                with the key attribute like a :class:`BaseDevice`, :class:`BaseNode` etc ...
        
        Example:
            
            ::
                
                > downloader = Downloader([mgr.motor1.substate, mgr.motor1.pos_actual, mgr.motor2.substate])
                > downloader.download()
                > m1_data = downloader.get_data_view(mgr.motor1.key) 
                > m1_data['pos_actual']
                3.9898
            
            ::
            
                > m1_data = downloader.get_data_view(mgr.motor1)
                # is equivalent to 
                > m1_data = DataView(downloader.data, mgr.motor1)
        """
        return DataView(self._data, prefix)        
    
    def clean_data(self) -> None:
        """ Remove to the .data dictionary all keys/value pairs corresponding to nodes not in the downloader queue
        
        Returns:
           n (int): The number of node/value pair removed 
        """
        d = self.data
        n = 0
        for n in list(d): # list(d) in order to avoid deletion on the iterator
            if not n in self._nodes:
                d.pop(n, None)
                n+=1
        return n


def reset(nodes):
    """ from a list of nodes reset the ones with a reset method defined 
    
    the reset method shall not take positional argument
    """
    for n in nodes:
        try:
            r = n.reset
        except AttributeError:
            pass
        else:
            r()

def download(nodes, data: Optional[Dict] = None) -> Union[list,None]:
    """ read node values from remotes in one call per remote
    
    Args:
        nodes (iterable):
             Iterable of nodes, like [mgr.motor1.stat.pos_actual, mgr.motor2.stat.pos_actual]
        
        data (dict, optional):
             This is mostlikely a dictionary, must define a __setitem__ method
             If given the function return None and update data in place. 
             If data is None the function return a list of values 
             
        
    Returns:
       None, or list : download(nodes) -> return list of values 
                       download(nodes, data) -> return None and update the input data dictionary
    
    Example:
    
    ::
        
        data = {n:n.get() for n in nodes}
    
    Is equivalent, but **much slower** than 
    
    :: 
        
        data = {}
        download(nodes)
        
    The latest is more efficient because only one call (per server) is done.
    
    data dictionary is optional, if not given values are returned in a list:
    
    ::
     
        pos, error = download([mgr.motor1.stat.pos_actual, mgr.motor1.stat.pos_error])
     
    
    """
    if data is None:
        data = {}
        nodes = tuple(nodes) # in case this is a generator  
        NodesReader(nodes).read(data)
        return [data[n] for n in nodes]
    else:    
        NodesReader(nodes).read(data)
        return None





from ._core_objects import *
from ._core_objects import _BaseObject
from .download import  Downloader, download, DataView
from .upload import upload, Uploader
from .wait import wait, Waiter
from .datamodel import (DataLink, BaseData, NodeVar, NodeVar_R, NodeVar_W,
                        NodeVar_RW, StaticVar, model_subset)

from ._misc.math_parser import DataEval, ExpEval
from .toolbox import *


try:
    import numpy
except ModuleNotFoundError:
    pass
else:
    from .np_nodes import *
    del numpy



from pydantic import BaseModel, ValidationError, Field
from pydantic.fields import ModelField
from typing import TypeVar, Generic, Any, Iterable, Dict, List, Type

ValType = TypeVar('ValType')

class NodeVar(Generic[ValType]):
    """
    A Field as NodeVar. Does not do validation by itself but it is used
    as an iddentifier of a node value
    """
    @classmethod
    def __get_validators__(cls):
        # one or more validators may be yielded which will be called in the
        # order to validate the input, each validator will receive as an input
        # the value returned from the previous validator
        yield cls.validate

    @classmethod
    def __modify_schema__(cls, field_schema):
        # __modify_schema__ should mutate the dict it receives in place,
        # the returned value will be ignored
        #field_schema.update()
        pass
        
    @classmethod
    def validate(cls, v, field: ModelField):
        if not field.sub_fields:
            # Generic parameters were not provided so we don't try to validate
            # them and just return the value as is
            return v
        if len(field.sub_fields)>1:
            raise ValidationError(['to many field NodeVar accep only one'], cls)
        val_f = field.sub_fields[0]
        errors = []
        
        valid_value, error = val_f.validate(v, {}, loc='value')
        if error:
            errors.append(error)
        if errors:
            raise ValidationError(errors, cls)
        # Validation passed without errors, return validated value
        return valid_value
    
    def __repr__(self):
        return f'{self.__class__.__name__}({super().__repr__()})'

class NodeVar_RW(NodeVar):
    """ Alias of :class:`NodeVar` """
    pass
class NodeVar_W(NodeVar):
    """ Write Only version of :class:`NodeVar`"""
    pass
    
class NodeVar_R(NodeVar):
    """ Read Only version of :class:`NodeVar` """
    pass

class StaticVar(Generic[ValType]):
    """
    A Field as StaticVar. Does not do validation by itself but it is used
    as an iddentifier of a static attribute of the input object of DataLink
    """
    @classmethod
    def __get_validators__(cls):
        # one or more validators may be yielded which will be called in the
        # order to validate the input, each validator will receive as an input
        # the value returned from the previous validator
        yield cls.validate

    @classmethod
    def __modify_schema__(cls, field_schema):
        # __modify_schema__ should mutate the dict it receives in place,
        # the returned value will be ignored
        #field_schema.update()
        pass
        
    @classmethod
    def validate(cls, v, field: ModelField):
        if not field.sub_fields:
            # Generic parameters were not provided so we don't try to validate
            # them and just return the value as is
            return v
        if len(field.sub_fields)>1:
            raise ValidationError(['to many field StaticVar accept only one'], cls)
        val_f = field.sub_fields[0]
        errors = []
        
        valid_value, error = val_f.validate(v, {}, loc='value')
        if error:
            errors.append(error)
        if errors:
            raise ValidationError(errors, cls)
        # Validation passed without errors, return validated value
        return valid_value
    
    def __repr__(self):
        return f'{self.__class__.__name__}({super().__repr__()})'

